# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import sys
import logging
from weakref import ref as weakref_ref
from pyomo.common.pyomo_typing import overload

from pyomo.common.deprecation import RenamedClass, deprecated
from pyomo.common.errors import TemplateExpressionError
from pyomo.common.enums import ObjectiveSense, minimize, maximize
from pyomo.common.log import is_debug_set
from pyomo.common.modeling import NOTSET
from pyomo.common.formatting import tabular_writer
from pyomo.common.timing import ConstructionTimer

from pyomo.core.expr.expr_common import _type_check_exception_arg
from pyomo.core.expr.numvalue import value
from pyomo.core.expr.template_expr import templatize_rule
from pyomo.core.base.component import ActiveComponentData, ModelComponentFactory
from pyomo.core.base.disable_methods import disable_methods
from pyomo.core.base.global_set import UnindexedComponent_index
from pyomo.core.base.indexed_component import (
    ActiveIndexedComponent,
    UnindexedComponent_set,
    rule_wrapper,
)
from pyomo.core.base.expression import NamedExpressionData
from pyomo.core.base.set import Set
from pyomo.core.base.initializer import (
    Initializer,
    IndexedCallInitializer,
    CountedCallInitializer,
)

logger = logging.getLogger('pyomo.core')

TEMPLATIZE_OBJECTIVES = False

_rule_returned_none_error = """Objective '%s': rule returned None.

Objective rules must return either a valid expression, numeric value, or
Objective.Skip.  The most common cause of this error is forgetting to
include the "return" statement at the end of your rule.
"""


def simple_objective_rule(rule):
    """
    This is a decorator that translates None into Objective.Skip.
    This supports a simpler syntax in objective rules, though these
    can be more difficult to debug when errors occur.

    Example use:

    .. code::

        @simple_objective_rule
        def O_rule(model, i, j):
            # ...

        model.o = Objective(rule=simple_objective_rule(...))

    """
    return rule_wrapper(rule, {None: Objective.Skip})


def simple_objectivelist_rule(rule):
    """
    This is a decorator that translates None into ObjectiveList.End.
    This supports a simpler syntax in objective rules, though these
    can be more difficult to debug when errors occur.

    Example use:

    .. code::

        @simple_objectivelist_rule
        def O_rule(model, i, j):
            # ...

        model.o = ObjectiveList(expr=simple_objectivelist_rule(...))

    """
    return rule_wrapper(rule, {None: ObjectiveList.End})


class ObjectiveData(NamedExpressionData, ActiveComponentData):
    """This class defines the data for a single objective.

    Note that this is a subclass of NumericValue to allow
    objectives to be used as part of expressions.

    Parameters
    ----------
    expr:
        The Pyomo expression stored in this objective.

    sense:
        The direction for this objective.

    component: Objective
        The Objective object that owns this data.

    Attributes
    ----------
    expr:
        The Pyomo expression for this objective

    """

    __slots__ = ("_args_", "_sense")

    def __init__(self, expr=None, sense=minimize, component=None):
        # Inlining NamedExpressionData.__init__
        self._args_ = (expr,)
        # Inlining ActiveComponentData.__init__
        self._component = weakref_ref(component) if (component is not None) else None
        self._index = NOTSET
        self._active = True
        self._sense = ObjectiveSense(sense)

    def is_minimizing(self):
        """Return True if this is a minimization objective."""
        return self.sense == minimize

    def set_value(self, expr):
        if expr is None:
            raise ValueError(_rule_returned_none_error % (self.name,))
        return super().set_value(expr)

    #
    # Abstract Interface
    #

    @property
    def sense(self):
        """Access sense (direction) of this objective."""
        return self._sense

    @sense.setter
    def sense(self, sense):
        """Set the sense (direction) of this objective."""
        self.set_sense(sense)

    def set_sense(self, sense):
        """Set the sense (direction) of this objective."""
        self._sense = ObjectiveSense(sense)


class _ObjectiveData(metaclass=RenamedClass):
    __renamed__new_class__ = ObjectiveData
    __renamed__version__ = '6.7.2'


class _GeneralObjectiveData(metaclass=RenamedClass):
    __renamed__new_class__ = ObjectiveData
    __renamed__version__ = '6.7.2'


class TemplateDataMixin:
    __slots__ = ()

    @property
    def args(self):
        # Note that it is faster to just generate the expression from
        # scratch than it is to clone it and replace the IndexTemplate objects
        self.set_value(self.parent_component().rule(self.parent_block(), self.index()))
        return self._args_

    def template_expr(self):
        return self._args_

    def set_value(self, expr):
        # Setting a value will convert this instance from a templatized
        # type to the original Data type (and call the original set_value()).
        #
        # Note: We assume that the templatized type is created by
        # inheriting (TemplateDataMixin, <original data class>), and
        # that this instance doesn't have additional multiple
        # inheritance that could re-order the MRO.
        self.__class__ = self.__class__.__mro__[
            self.__class__.__mro__.index(TemplateDataMixin) + 1
        ]
        return self.set_value(expr)


class TemplateObjectiveData(TemplateDataMixin, ObjectiveData):
    __slots__ = ()

    def __init__(self, template_info, component, index, sense):
        #
        # These lines represent in-lining of the
        # following constructors:
        #   - ObjectiveData
        #   - ActiveComponentData
        #   - ComponentData
        self._component = component
        self._active = True
        self._index = index
        self._args_ = template_info
        self._sense = sense


@ModelComponentFactory.register("Expressions that are minimized or maximized.")
class Objective(ActiveIndexedComponent):
    """
    This modeling component defines an objective expression.

    Note that this is a subclass of NumericValue to allow
    objectives to be used as part of expressions.

    Constructor arguments:
        expr
            A Pyomo expression for this objective
        rule
            A function that is used to construct objective expressions
        sense
            Indicate whether minimizing (the default) or maximizing
        name
            A name for this component
        doc
            A text string describing this component

    Public class attributes:
        doc
            A text string describing this component
        name
            A name for this component
        active
            A boolean that is true if this component will be used to construct
            a model instance
        rule
            The rule used to initialize the objective(s)
        sense
            The objective sense

    Private class attributes:
        _constructed
            A boolean that is true if this component has been constructed
        _data
            A dictionary from the index set to component data objects
        _index
            The set of valid indices
        _model
            A weakref to the model that owns this component
        _parent
            A weakref to the parent block that owns this component
        _type
            The class type for the derived subclass
    """

    _ComponentDataClass = ObjectiveData
    NoObjective = ActiveIndexedComponent.Skip

    def __new__(cls, *args, **kwds):
        if cls != Objective:
            return super().__new__(cls)
        if not args or (args[0] is UnindexedComponent_set and len(args) == 1):
            return super().__new__(AbstractScalarObjective)
        else:
            return super().__new__(IndexedObjective)

    @overload
    def __init__(
        self, *indexes, expr=None, rule=None, sense=minimize, name=None, doc=None
    ): ...

    def __init__(self, *args, **kwargs):
        _sense = kwargs.pop('sense', minimize)
        _init = self._pop_from_kwargs('Objective', kwargs, ('rule', 'expr'), None)

        kwargs.setdefault('ctype', Objective)
        ActiveIndexedComponent.__init__(self, *args, **kwargs)

        self._rule = Initializer(_init)
        self._init_sense = Initializer(_sense)

    def construct(self, data=None):
        """
        Construct the expression(s) for this objective.
        """
        if self._constructed:
            return
        self._constructed = True

        timer = ConstructionTimer(self)
        if is_debug_set(logger):
            logger.debug("Constructing objective %s" % (self.name))

        if self._anonymous_sets is not None:
            for _set in self._anonymous_sets:
                _set.construct()

        rule = self._rule
        try:
            # We do not (currently) accept data for constructing Objectives
            index = None
            assert data is None

            if rule is None:
                # If there is no rule, then we are immediately done.
                return

            if rule.constant() and self.is_indexed():
                raise IndexError(
                    "Objective '%s': Cannot initialize multiple indices "
                    "of an objective with a single expression" % (self.name,)
                )

            block = self.parent_block()
            if rule.contains_indices():
                # The index is coming in externally; we need to validate it
                for index in rule.indices():
                    ans = self.__setitem__(index, rule(block, index))
                    if ans is not None:
                        self[index].set_sense(self._init_sense(block, index))
            elif not self.index_set().isfinite():
                # If the index is not finite, then we cannot iterate
                # over it.  Since the rule doesn't provide explicit
                # indices, then there is nothing we can do (the
                # assumption is that the user will trigger specific
                # indices to be created at a later time).
                pass
            else:
                if TEMPLATIZE_OBJECTIVES:
                    try:
                        template_info = templatize_rule(block, rule, self.index_set())
                        if self.is_indexed():
                            comp = weakref_ref(self)
                            self._data = {
                                idx: TemplateObjectiveData(
                                    template_info,
                                    comp,
                                    idx,
                                    self._init_sense(block, index),
                                )
                                for idx in self.index_set()
                            }
                        else:
                            assert self.__class__ is ScalarObjective
                            self.__class__ = TemplateScalarObjective
                            self._args_ = template_info
                            self._data = {None: self}
                            self.set_sense(self._init_sense(self, self.index()))
                        return
                    except TemplateExpressionError:
                        pass

                # Bypass the index validation and create the member directly
                for index in self.index_set():
                    ans = self._setitem_when_not_present(index, rule(block, index))
                    if ans is not None:
                        ans.set_sense(self._init_sense(block, index))
        except Exception:
            err = sys.exc_info()[1]
            logger.error(
                "Rule failed when generating expression for "
                "Objective %s with index %s:\n%s: %s"
                % (self.name, str(index), type(err).__name__, err)
            )
            raise
        finally:
            timer.report()

    def _getitem_when_not_present(self, index):
        if self._rule is None:
            raise KeyError(index)

        block = self.parent_block()
        obj = self._setitem_when_not_present(index, self._rule(block, index))
        if obj is None:
            raise KeyError(index)
        obj.set_sense(self._init_sense(block, index))

        return obj

    def _pprint(self):
        """
        Return data that will be printed for this component.
        """
        return (
            [
                ("Size", len(self)),
                ("Index", self._index_set if self.is_indexed() else None),
                ("Active", self.active),
            ],
            self.items,
            ("Active", "Sense", "Expression"),
            lambda k, v: [v.active, v.sense, v.expr],
        )

    @property
    def rule(self):
        return self._rule

    @rule.setter
    @deprecated(
        f"The 'Objective.rule' attribute will be made "
        "read-only in a future Pyomo release.",
        version='6.9.3',
        remove_in='6.11',
    )
    def rule(self, rule):
        self._rule = rule

    def display(self, prefix="", ostream=None):
        """Provide a verbose display of this object"""
        if not self.active:
            return
        tab = "    "
        if ostream is None:
            ostream = sys.stdout
        ostream.write(prefix + self.local_name + " : ")
        ostream.write(
            ", ".join(
                "%s=%s" % (k, v)
                for k, v in [
                    ("Size", len(self)),
                    ("Index", self._index_set if self.is_indexed() else None),
                    ("Active", self.active),
                ]
            )
        )

        ostream.write("\n")
        tabular_writer(
            ostream,
            prefix + tab,
            ((k, v) for k, v in self._data.items() if v.active),
            ("Active", "Value"),
            lambda k, v: [v.active, value(v)],
        )


class ScalarObjective(ObjectiveData, Objective):
    """
    ScalarObjective is the implementation representing a single,
    non-indexed objective.
    """

    def __init__(self, *args, **kwd):
        ObjectiveData.__init__(self, expr=None, component=self)
        Objective.__init__(self, *args, **kwd)
        self._index = UnindexedComponent_index

    def __call__(self, exception=NOTSET):
        exception = _type_check_exception_arg(self, exception)
        if not self._data:
            raise ValueError(
                "Evaluating the expression of ScalarObjective "
                "'%s' before the Objective has been assigned "
                "a sense or expression (there is currently "
                "no value to return)." % (self.name)
            )
        return super().__call__(exception)

    #
    # Singleton objectives are strange in that we want them to be
    # both be constructed but have len() == 0 when not initialized with
    # anything (at least according to the unit tests that are
    # currently in place). So during initialization only, we will
    # treat them as "indexed" objects where things like
    # Objective.Skip are managed. But after that they will behave
    # like ObjectiveData objects where set_value does not handle
    # Objective.Skip but expects a valid expression or None
    #

    @property
    def sense(self):
        """Access sense (direction) of this objective."""
        if self._constructed:
            if len(self._data) == 0:
                raise ValueError(
                    "Accessing the sense of ScalarObjective "
                    "'%s' before the Objective has been assigned "
                    "a sense or expression (there is currently "
                    "no value to return)." % (self.name)
                )
            return ObjectiveData.sense.fget(self)
        raise ValueError(
            "Accessing the sense of objective '%s' "
            "before the Objective has been constructed (there "
            "is currently no value to return)." % (self.name)
        )

    @sense.setter
    def sense(self, sense):
        """Set the sense (direction) of this objective."""
        self.set_sense(sense)

    def clear(self):
        self._data = {}

    def set_value(self, expr):
        """Set the expression of this objective."""
        if not self._data:
            self._data[None] = self
        return super().set_value(expr)

    def set_sense(self, sense):
        """Set the sense (direction) of this objective."""
        if len(self._data) == 0:
            self._data[None] = self
        return ObjectiveData.set_sense(self, sense)

    #
    # Leaving this method for backward compatibility reasons.
    # (probably should be removed)
    #
    def add(self, index, expr):
        """Add an expression with a given index."""
        if index is not None:
            raise ValueError(
                "ScalarObjective object '%s' does not accept "
                "index values other than None. Invalid value: %s" % (self.name, index)
            )
        self.set_value(expr)
        return self


class SimpleObjective(metaclass=RenamedClass):
    __renamed__new_class__ = ScalarObjective
    __renamed__version__ = '6.0'


@disable_methods({'__call__', 'add', 'expr', 'sense', 'set_sense', 'set_value'})
class AbstractScalarObjective(ScalarObjective):
    pass


class TemplateScalarObjective(TemplateDataMixin, ScalarObjective):
    pass


class IndexedObjective(Objective):
    #
    # Leaving this method for backward compatibility reasons
    #
    # Note: Beginning after Pyomo 5.2 this method will now validate that
    # the index is in the underlying index set (through 5.2 the index
    # was not checked).
    #
    def add(self, index, expr):
        """Add an objective with a given index."""
        return self.__setitem__(index, expr)


@ModelComponentFactory.register("A list of objective expressions.")
class ObjectiveList(IndexedObjective):
    """
    An objective component that represents a list of objectives.
    Objectives can be indexed by their index, but when they are added
    an index value is not specified.
    """

    class End:
        pass

    def __init__(self, **kwargs):
        """Constructor"""
        if 'expr' in kwargs:
            raise ValueError("ObjectiveList does not accept the 'expr' keyword")
        _rule = kwargs.pop('rule', None)
        self._starting_index = kwargs.pop('starting_index', 1)

        super().__init__(Set(dimen=1), **kwargs)

        self._rule = Initializer(_rule, allow_generators=True)
        # HACK to make the "counted call" syntax work.  We wait until
        # after the base class is set up so that is_indexed() is
        # reliable.
        if self._rule is not None and type(self._rule) is IndexedCallInitializer:
            self._rule = CountedCallInitializer(self, self._rule, self._starting_index)

    def construct(self, data=None):
        """
        Construct the expression(s) for this objective.
        """
        if self._constructed:
            return
        self._constructed = True

        if is_debug_set(logger):
            logger.debug("Constructing objective list %s" % (self.name))

        if self._anonymous_sets is not None:
            for _set in self._anonymous_sets:
                _set.construct()

        if self._rule is not None:
            _rule = self._rule(self.parent_block(), ())
            for cc in iter(_rule):
                if cc is ObjectiveList.End:
                    break
                if cc is Objective.Skip:
                    continue
                self.add(cc, sense=self._init_sense)

    def add(self, expr, sense=minimize):
        """Add an objective to the list."""
        next_idx = len(self._index_set) + self._starting_index
        self._index_set.add(next_idx)
        ans = self.__setitem__(next_idx, expr)
        if ans is not None:
            if sense not in {minimize, maximize}:
                sense = sense(self.parent_block(), next_idx)
            ans.set_sense(sense)
        return ans



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from __future__ import annotations
import sys
import types
import logging
from typing import Union, Type
from weakref import ref as weakref_ref

from pyomo.common.autoslots import AutoSlots
from pyomo.common.deprecation import deprecated, deprecation_warning, RenamedClass
from pyomo.common.log import is_debug_set
from pyomo.common.modeling import NOTSET
from pyomo.common.numeric_types import native_types, value as expr_value
from pyomo.common.pyomo_typing import overload
from pyomo.common.timing import ConstructionTimer
from pyomo.core.expr.expr_common import _type_check_exception_arg
from pyomo.core.expr.numvalue import NumericValue
from pyomo.core.base.component import ComponentData, ModelComponentFactory
from pyomo.core.base.enums import SortComponents
from pyomo.core.base.global_set import UnindexedComponent_index
from pyomo.core.base.indexed_component import (
    IndexedComponent,
    UnindexedComponent_set,
    IndexedComponent_NDArrayMixin,
)
from pyomo.core.base.initializer import Initializer, PartialInitializer
from pyomo.core.base.misc import apply_indexed_rule, apply_parameterized_indexed_rule
from pyomo.core.base.set import Reals, _AnySet, SetInitializer
from pyomo.core.base.units_container import units
from pyomo.core.expr import GetItemExpression

logger = logging.getLogger('pyomo.core')


def _placeholder_rule(*args, **kwargs):
    pass


def _raise_modifying_immutable_error(obj, index):
    if obj.is_indexed():
        name = "%s[%s]" % (obj.name, index)
    else:
        name = obj.name
    raise TypeError(
        "Attempting to set the value of the immutable parameter "
        "%s after the parameter has been constructed.  If you intend "
        "to change the value of this parameter dynamically, please "
        "declare the parameter as mutable [i.e., Param(mutable=True)]" % (name,)
    )


class _ImplicitAny(_AnySet):
    """An Any that issues a deprecation warning for non-Real values.

    This is a helper class to implement the deprecation warnings for the
    change of Param's implicit domain from Any to Reals.

    """

    __slots__ = ('_owner',)
    __autoslot_mappers__ = {'_owner': AutoSlots.weakref_mapper}

    def __new__(cls, **kwargs):
        # Strip off owner / kwargs before calling base __new__
        return super().__new__(cls)

    def __init__(self, owner, **kwargs):
        self._owner = weakref_ref(owner)
        super().__init__(**kwargs)
        self._component = weakref_ref(self)
        self.construct()
        # Because this is a "global set", we need to define the _bounds
        # and _interval fields
        object.__setattr__(self, '_parent', None)
        self._bounds = (None, None)
        self._interval = (None, None, None)

    def __contains__(self, val):
        if val not in Reals:
            if self._owner is None or self._owner() is None:
                name = 'Unknown'
            else:
                name = self._owner().name
            deprecation_warning(
                f"Param '{name}' declared with an implicit domain of 'Any'. "
                "The default domain for Param objects is 'Any'.  However, "
                "we will be changing that default to 'Reals' in the "
                "future.  If you really intend the domain of this Param"
                "to be 'Any', you can suppress this warning by explicitly "
                "specifying 'within=Any' to the Param constructor.",
                version='5.6.9',
                remove_in='6.0',
            )
        return True

    # This should "mock up" a global set, so the "name" should always be
    # the local name (without block scope)
    def getname(self, fully_qualified=False, name_buffer=None, relative_to=None):
        return super().getname(False, name_buffer, relative_to)

    # The parent tracks the parent of the owner.  We can't set it
    # directly here because the owner has not been assigned to a block
    # when we create the _ImplicitAny
    @property
    def _parent(self):
        if self._owner is None or self._owner() is None:
            return None
        return self._owner()._parent

    # This is not settable.  However the base classes assume that it is,
    # so we need to define the setter and just ignore the incoming value
    @_parent.setter
    def _parent(self, val):
        pass


class ParamData(ComponentData, NumericValue):
    """
    This class defines the data for a mutable parameter.

    Constructor Arguments:
        owner       The Param object that owns this data.
        value       The value of this parameter.

    Public Class Attributes:
        value       The numeric value of this variable.
    """

    __slots__ = ('_value',)

    def __init__(self, component):
        #
        # The following is equivalent to calling
        # the base ComponentData constructor.
        #
        self._component = weakref_ref(component)
        self._index = NOTSET
        #
        # The following is equivalent to calling the
        # base NumericValue constructor.
        #
        self._value = Param.NoValue

    # Note: because NONE of the slots on this class need to be edited,
    # we don't need to implement a specialized __setstate__ method.

    def clear(self):
        """Clear the data in this component"""
        self._value = Param.NoValue

    # FIXME: ComponentData need to have pointers to their index to make
    # operations like validation efficient.  As it stands now, if
    # set_value is called without specifying an index, this call
    # involves a linear scan of the _data dict.
    def set_value(self, value, idx=NOTSET):
        """Set the value of this ParamData object, performing unit conversion
        and validation as necessary.

        """
        #
        # If this param has units, then we need to check the incoming
        # value and see if it is "units compatible".  We only need to
        # check here in set_value, because all united Params are
        # required to be mutable.
        #
        _comp = self.parent_component()
        if value.__class__ in native_types:
            # TODO: warn/error: check if this Param has units: assigning
            # a dimensionless value to a united param should be an error
            pass
        elif _comp._units is not None:
            _src_magnitude = expr_value(value)
            # Note: expr_value() could have just registered a new numeric type
            if value.__class__ in native_types:
                value = _src_magnitude
            else:
                _src_units = units.get_units(value)
                value = units.convert_value(
                    num_value=_src_magnitude,
                    from_units=_src_units,
                    to_units=_comp._units,
                )
        # FIXME: we should call value() here [to ensure types get
        # registered], but doing so breaks non-numeric Params (which we
        # allow).  The real fix will be to follow the precedent from
        # GetItemExpression and have separate types based on which
        # expression "system" the Param should participate in (numeric,
        # logical, or structural).
        #
        # else:
        #     value = expr_value(value)

        old_value, self._value = self._value, value
        try:
            _comp._validate_value(idx, value, data=self)
        except:
            self._value = old_value
            raise

    def __call__(self, exception=NOTSET):
        """
        Return the value of this object.
        """
        exception = _type_check_exception_arg(self, exception)

        if self._value is Param.NoValue:
            if exception:
                raise ValueError(
                    "Error evaluating Param value (%s):\n\tThe Param value is "
                    "currently set to an invalid value.  This is\n\ttypically "
                    "from a scalar Param or mutable Indexed Param without\n"
                    "\tan initial or default value." % (self.name,)
                )
            else:
                return None
        return self._value

    @property
    def value(self):
        """Return the value for this variable."""
        return self()

    @value.setter
    def value(self, val):
        """Set the value for this variable."""
        self.set_value(val)

    def get_units(self):
        """Return the units for this ParamData"""
        return self.parent_component()._units

    def is_fixed(self):
        """
        Returns True because this value is fixed.
        """
        return True

    def is_constant(self):
        """
        Returns False because this is not a constant in an expression.
        """
        return False

    def is_parameter_type(self):
        """
        Returns True because this is a parameter object.
        """
        return True

    def _compute_polynomial_degree(self, result):
        """
        Returns 0 because this object can never reference variables.
        """
        return 0


class _ParamData(metaclass=RenamedClass):
    __renamed__new_class__ = ParamData
    __renamed__version__ = '6.7.2'


@ModelComponentFactory.register(
    "Parameter data that is used to define a model instance."
)
class Param(IndexedComponent, IndexedComponent_NDArrayMixin):
    """
    A parameter value, which may be defined over an index.

    Constructor Arguments:
        domain
            A set that defines the type of values that each parameter must be.
        within
            A set that defines the type of values that each parameter must be.
        validate
            A rule for validating this parameter w.r.t. data that exists in
            the model
        default
            A scalar, rule, or dictionary that defines default values for
            this parameter
        initialize
            A dictionary or rule for setting up this parameter with existing
            model data
        unit: pyomo unit expression
            An expression containing the units for the parameter
        mutable: `boolean`
            Flag indicating if the value of the parameter may change between
            calls to a solver. Defaults to `False`
        name
            Name for this component.
        doc
            Text describing this component.
    """

    DefaultMutable = False
    _ComponentDataClass = ParamData

    class NoValue:
        """A dummy type that is pickle-safe that we can use as the default
        value for Params to indicate that no valid value is present."""

        pass

    @overload
    def __new__(cls: Type[ScalarParam], *args, **kwds) -> ScalarParam: ...

    @overload
    def __new__(cls: Type[IndexedParam], *args, **kwds) -> IndexedParam: ...

    @overload
    def __new__(
        cls: Type[Param], *args, **kwds
    ) -> Union[ScalarParam, IndexedParam]: ...

    def __new__(cls, *args, **kwds):
        if cls != Param:
            return super(Param, cls).__new__(cls)
        if not args or (args[0] is UnindexedComponent_set and len(args) == 1):
            return super(Param, cls).__new__(ScalarParam)
        else:
            return super(Param, cls).__new__(IndexedParam)

    @overload
    def __init__(
        self,
        *indexes,
        rule=NOTSET,
        initialize=NOTSET,
        domain=None,
        within=None,
        validate=None,
        mutable=False,
        default=NoValue,
        initialize_as_dense=False,
        units=None,
        name=None,
        doc=None,
    ): ...

    def __init__(self, *args, **kwd):
        _init = self._pop_from_kwargs('Param', kwd, ('rule', 'initialize'), NOTSET)
        _domain_rule = self._pop_from_kwargs('Param', kwd, ('domain', 'within'))
        self._validate = kwd.pop('validate', None)
        self._mutable = kwd.pop('mutable', None)
        self._default_val = kwd.pop('default', Param.NoValue)
        self._dense_initialize = kwd.pop('initialize_as_dense', False)
        self._units = kwd.pop('units', None)

        if self._mutable is None:
            if self._units is None:
                self._mutable = Param.DefaultMutable
            else:
                # Params with units *must* be mutable, so that
                # expression simplification does not remove units from
                # the expression.
                self._mutable = True
        if _init is not NOTSET:
            # We need a placeholder rule on the Param because the base
            # class will wrap it to pass in any unrecognized keyword
            # arguments.  We can't just pass the actual rule because
            # we want to use is_indexed() to change how we process the rule.
            self._rule = _placeholder_rule
        else:
            self._rule = None

        kwd.setdefault('ctype', Param)
        IndexedComponent.__init__(self, *args, **kwd)

        # We don't support per-index param domains, so we only need to
        # support constant initializers.
        # (after IndexedComponent.__init__ so we can call parent_block())
        if _domain_rule is None:
            self.domain = _ImplicitAny(owner=self, name='Any')
        else:
            self.domain = SetInitializer(_domain_rule)(self.parent_block(), None, self)
        # After IndexedComponent.__init__ so we can call is_indexed().
        _rule = Initializer(
            _init,
            treat_sequences_as_mappings=self.is_indexed(),
            arg_not_specified=NOTSET,
        )
        if self._rule.__class__ is PartialInitializer:
            # Replace the _placeholder_rule with the user-specified rule
            self._rule = PartialInitializer(
                _rule, *self._rule._fcn.args, **self._rule._fcn.keywords
            )
        else:
            self._rule = _rule

    def __len__(self):
        """
        Return the number of component data objects stored by this
        component.  If a default value is specified, then the
        length equals the number of items in the component index.
        """
        if self._default_val is Param.NoValue:
            return len(self._data)
        return len(self._index_set)

    def __contains__(self, idx):
        """
        Return true if the index is in the dictionary.  If the default value
        is specified, then all members of the component index are valid.
        """
        if self._default_val is Param.NoValue:
            return idx in self._data
        return idx in self._index_set

    # We do not need to override keys(), as the __len__ override will
    # cause the base class keys() to correctly correctly handle default
    # values
    # def keys(self, sort=None):

    @property
    def mutable(self):
        return self._mutable

    def get_units(self):
        """Return the units for this Param"""
        return self._units

    #
    # These are "sparse equivalent" access / iteration methods that
    # only loop over the defined data.
    #

    def sparse_keys(self, sort=SortComponents.UNSORTED):
        """Return a list of keys in the defined parameters"""
        try:
            # Temporarily remove the default value so that len(self) ==
            # len(self._dict).  This will cause the base class
            # implementation of keys() to only return values from
            # self._data:
            tmp = self._default_val
            self._default_val = Param.NoValue
            return self.keys(sort)
        finally:
            self._default_val = tmp

    def sparse_values(self, sort=SortComponents.UNSORTED):
        """Return a list of the defined param data objects"""
        # Implementing things this way for consistency with items() (and
        # so that any changes in the base class implementation are
        # picked up here, too):
        try:
            tmp = self._default_val
            self._default_val = Param.NoValue
            return self.values(sort)
        finally:
            self._default_val = tmp

    def sparse_items(self, sort=SortComponents.UNSORTED):
        """Return a list (index,data) tuples for defined parameters"""
        # The base class implements special handling for references.
        # Instead of reimplementing that here, we will follow the
        # pattern used for sparse_keys (and get len() to "lie")
        try:
            tmp = self._default_val
            self._default_val = Param.NoValue
            return self.items(sort)
        finally:
            self._default_val = tmp

    @deprecated(
        "The sparse_iterkeys method is deprecated.  Use sparse_keys()",
        # This should have been deprecated when we dropped Python 2.7
        version='6.10.0',
    )
    def sparse_iterkeys(self):
        """Return an iterator for the keys in the defined parameters"""
        return self.sparse_keys()

    @deprecated(
        "The sparse_itervalues method is deprecated.  Use sparse_values()",
        # This should have been deprecated when we dropped Python 2.7
        version='6.10.0',
    )
    def sparse_itervalues(self):
        """Return an iterator for the defined param data objects"""
        return self.sparse_values()

    @deprecated(
        "The sparse_iteritems method is deprecated.  Use sparse_items()",
        # This should have been deprecated when we dropped Python 2.7
        version='6.10.0',
    )
    def sparse_iteritems(self):
        """Return an iterator of (index,data) tuples for defined parameters"""
        return self.sparse_items()

    def extract_values(self):
        """
        A utility to extract all index-value pairs defined for this
        parameter, returned as a dictionary.

        This method is useful in contexts where key iteration and
        repeated __getitem__ calls are too expensive to extract
        the contents of a parameter.
        """
        if self._mutable:
            # The parameter is mutable so parameter data are ParamData
            # types.  We need to evaluate the ParamData back to POD
            # (numeric) data when creating the result.
            return {key: param_data() for key, param_data in self.items()}
        elif not self.is_indexed():
            # The scalar could be defined (in which case items() will
            # return the ScalarParam), OR it could be defined by a
            # default value (in which case items() will return the
            # actual numeric value).  To cover both cases we will use
            # value():
            return {key: expr_value(param_data) for key, param_data in self.items()}
        else:
            # The parameter is not mutable, so iteritems() can be
            # converted into a dictionary containing parameter values.
            return dict(self.items())

    def extract_values_sparse(self):
        """
        A utility to extract all index-value pairs defined with non-default
        values, returned as a dictionary.

        This method is useful in contexts where key iteration and
        repeated __getitem__ calls are too expensive to extract
        the contents of a parameter.
        """
        if self._mutable or not self.is_indexed():
            # The parameter is mutable so parameter data are ParamData
            # types.  We need to evaluate the ParamData back to POD
            # (numeric) data when creating the result.
            #
            # Note that if this is a scalar, sparse_items will return
            # the ScalarParam only if it is explicitly defined (in which
            # case it will still be evaluatable by calling it).
            # ScalarParams whose value comes from the default are not
            # returned by sparse_items()
            return {key: param_data() for key, param_data in self.sparse_items()}
        else:
            # The parameter is not mutable, so sparse_items() can be
            # converted into a dictionary containing parameter values.
            return dict(self.sparse_items())

    def store_values(self, new_values, check=True):
        """
        A utility to update a Param with a dictionary or scalar.

        If check=True, then both the index and value
        are checked through the __getitem__ method.  Using check=False
        should only be used by developers!
        """
        if not self._mutable:
            _raise_modifying_immutable_error(self, '*')
        #
        _srcType = type(new_values)
        _isDict = _srcType is dict or (
            hasattr(_srcType, '__getitem__')
            and not isinstance(new_values, NumericValue)
        )
        #
        if check:
            if _isDict:
                for index, new_value in new_values.items():
                    self[index] = new_value
            else:
                for index in self._index_set:
                    self[index] = new_values
            return
        #
        # The argument check is False, so we bypass almost all of the
        # Param logic for ensuring data integrity.
        #
        if self.is_indexed():
            if _isDict:
                # It is possible that the Param is sparse and that the
                # index is not already in the _data dict.  As these
                # cases are rare, we will recover from the exception
                # instead of incurring the penalty of checking.
                for index, new_value in new_values.items():
                    if index not in self._data:
                        self._data[index] = ParamData(self)
                    self._data[index]._value = new_value
            else:
                # For scalars, we will choose an approach based on
                # how "dense" the Param is
                if not self._data:  # empty
                    for index in self._index_set:
                        p = self._data[index] = ParamData(self)
                        p._value = new_values
                elif len(self._data) == len(self._index_set):
                    for index in self._index_set:
                        self._data[index]._value = new_values
                else:
                    for index in self._index_set:
                        if index not in self._data:
                            self._data[index] = ParamData(self)
                        self._data[index]._value = new_values
        else:
            #
            # Initialize a scalar
            #
            if _isDict:
                if None not in new_values:
                    raise RuntimeError(
                        "Cannot store value for scalar Param %s:\n\tNo value "
                        "with index None in the new values dict." % (self.name,)
                    )
                new_values = new_values[None]
            # scalars have to be handled differently
            self[None] = new_values

    def set_default(self, val):
        """
        Perform error checks and then set the default value for this parameter.

        NOTE: this test will not validate the value of function return values.
        """
        if (
            self._constructed
            and val is not Param.NoValue
            and type(val) in native_types
            and val not in self.domain
        ):
            raise ValueError(
                "Default value (%s) is not valid for Param %s domain %s"
                % (str(val), self.name, self.domain.name)
            )
        self._default_val = val

    def default(self):
        """
        Return the value of the parameter default.

        Possible values:
            Param.NoValue
                No default value is provided.
            Numeric
                A constant value that is the default value for all undefined
                parameters.
            Function
                f(model, i) returns the value for the default value for
                parameter i
        """
        return self._default_val

    def _getitem_when_not_present(self, index):
        """
        Returns the default component data value
        """
        #
        # Local values
        #
        val = self._default_val
        if val is Param.NoValue:
            # We should allow the creation of mutable params without
            # a default value, as long as *solving* a model without
            # reasonable values produces an informative error.
            if self._mutable:
                # Note: ParamData defaults to Param.NoValue
                if self.is_indexed():
                    ans = self._data[index] = ParamData(self)
                else:
                    ans = self._data[index] = self
                ans._index = index
                return ans
            if self.is_indexed():
                idx_str = '%s[%s]' % (self.name, index)
            else:
                idx_str = '%s' % (self.name,)
            raise ValueError(
                "Error retrieving immutable Param value (%s):\n\tThe Param "
                "value is undefined and no default value is specified." % (idx_str,)
            )

        _default_type = type(val)
        _check_value_domain = True
        if _default_type in native_types:
            #
            # The set_default() method validates the domain of native types, so
            # we can skip the check on the value domain.
            #
            _check_value_domain = False
        elif _default_type is types.FunctionType:
            val = apply_indexed_rule(self, val, self.parent_block(), index)
        elif hasattr(val, '__getitem__') and (
            not isinstance(val, NumericValue) or val.is_indexed()
        ):
            # Things that look like Dictionaries should be allowable.  This
            # includes other IndexedComponent objects.
            val = val[index]
        else:
            # this is something simple like a non-indexed component
            pass

        #
        # If the user wants to validate values, we need to validate the
        # default value as well. For Mutable Params, this is easy:
        # _setitem_impl will inject the value into _data and
        # then call validate.
        #
        if self._mutable:
            return self._setitem_when_not_present(index, val)
        #
        # For immutable params, we never inject the default into the data
        # dictionary.  This will break validation, as the validation rule is
        # allowed to assume the data is already present (actually, it will
        # die on infinite recursion, as Param.__getitem__() will re-call
        # _getitem_when_not_present).
        #
        # So, we will do something very inefficient: if we are
        # validating, we will inject the value into the dictionary,
        # call validate, and remove it.
        #
        if _check_value_domain or self._validate:
            try:
                self._data[index] = val
                self._validate_value(index, val, _check_value_domain)
            finally:
                del self._data[index]

        return val

    def _setitem_impl(self, index, obj, value):
        """The __setitem__ method performs significant validation around the
        input indices, particularly when the index value is new.  In
        various contexts, we don't need to incur this overhead
        (e.g. during initialization).  The _setitem_impl
        assumes the input value is in the set native_types

        """
        #
        # We need to ensure that users don't override the value for immutable
        # parameters.
        #
        if self._constructed and not self._mutable:
            _raise_modifying_immutable_error(self, index)
        #
        # Params should contain *values*.  Note that if we just call
        # value(), then that forces the value to be a numeric value.
        # Notably, we allow Params with domain==Any to hold strings, tuples,
        # etc.  The following lets us use NumericValues to initialize
        # Params, but is optimized to check for "known" native types to
        # bypass a potentially expensive isinstance()==False call.
        #
        if value.__class__ not in native_types:
            if isinstance(value, NumericValue):
                value = value()
        #
        # Set the value depending on the type of param value.
        #
        if self._mutable:
            obj.set_value(value, index)
            return obj
        else:
            old_value, self._data[index] = self._data[index], value
            # Because we do not have a ParamData, we cannot rely on the
            # validation that occurs in ParamData.set_value()
            try:
                self._validate_value(index, value)
                return value
            except:
                self._data[index] = old_value
                raise

    def _setitem_when_not_present(self, index, value, _check_domain=True):
        #
        # We need to ensure that users don't override the value for immutable
        # parameters.
        #
        if self._constructed and not self._mutable:
            _raise_modifying_immutable_error(self, index)
        #
        # Params should contain *values*.  Note that if we just call
        # value(), then that forces the value to be a numeric value.
        # Notably, we allow Params with domain==Any to hold strings, tuples,
        # etc.  The following lets us use NumericValues to initialize
        # Params, but is optimized to check for "known" native types to
        # bypass a potentially expensive isinstance()==False call.
        #
        if value.__class__ not in native_types:
            if isinstance(value, NumericValue):
                value = value()

        #
        # Set the value depending on the type of param value.
        #
        try:
            if index is None and not self.is_indexed():
                self._data[None] = self
                self.set_value(value, index)
                self._index = UnindexedComponent_index
                return self
            elif self._mutable:
                obj = self._data[index] = ParamData(self)
                obj.set_value(value, index)
                obj._index = index
                return obj
            else:
                self._data[index] = value
                # Because we do not have a ParamData, we cannot rely on the
                # validation that occurs in ParamData.set_value()
                self._validate_value(index, value, _check_domain)
                return value
        except:
            del self._data[index]
            raise

    def _validate_value(self, index, value, validate_domain=True, data=None):
        """
        Validate a given input/value pair.
        """
        #
        # Check if the value is valid within the current domain
        #
        if validate_domain and not value in self.domain:
            if index is NOTSET:
                index = data.index()
            raise ValueError(
                "Invalid parameter value: %s[%s] = '%s', value type=%s.\n"
                "\tValue not in parameter domain %s"
                % (self.name, index, value, type(value), self.domain.name)
            )
        if self._validate:
            if index is NOTSET:
                index = data.index()
            valid = apply_parameterized_indexed_rule(
                self, self._validate, self.parent_block(), value, index
            )
            if not valid:
                raise ValueError(
                    "Invalid parameter value: %s[%s] = '%s', value type=%s.\n"
                    "\tValue failed parameter validation rule"
                    % (self.name, index, value, type(value))
                )

    def construct(self, data=None):
        """
        Initialize this component.

        A parameter is constructed using the initial data or
        the data loaded from an external source.  We first
        set all the values based on self._rule, and then
        allow the data dictionary to overwrite anything.

        Note that we allow an undefined Param value to be
        constructed.  We throw an exception if a user tries
        to use an uninitialized Param.
        """
        if self._constructed:
            return

        timer = ConstructionTimer(self)
        if is_debug_set(logger):  # pragma:nocover
            logger.debug(
                "Constructing Param, name=%s, from data=%s" % (self.name, str(data))
            )

        if self._units is not None:
            self._units = units.get_units(self._units)
            if not self._mutable:
                logger.warning(
                    "Params with units must be mutable.  "
                    f"Converting Param '{self.name}' to mutable."
                )
                self._mutable = True

        if self._anonymous_sets is not None:
            for _set in self._anonymous_sets:
                _set.construct()

        try:
            #
            # If the default value is a simple type, we check it versus
            # the domain.
            #
            val = self._default_val
            if (
                val is not Param.NoValue
                and type(val) in native_types
                and val not in self.domain
            ):
                raise ValueError(
                    "Default value (%s) is not valid for Param %s domain %s"
                    % (str(val), self.name, self.domain.name)
                )
            #
            # Flag that we are in the "during construction" phase
            #
            self._constructed = None
            #
            # Step #1: initialize data from rule value
            #
            self._construct_from_rule_using_setitem()
            #
            # Step #2: allow any user-specified (external) data to override
            # the initialization
            #
            if data is not None:
                try:
                    data_items = data.items()
                except AttributeError:
                    raise ValueError(
                        "Attempting to initialize parameter=%s with data=%s.\n"
                        "\tData type is not a mapping type, and a Mapping is "
                        "expected." % (self.name, str(data))
                    )
            else:
                data_items = iter(())

            try:
                for key, val in data_items:
                    self._setitem_when_not_present(self._validate_index(key), val)
            except:
                msg = sys.exc_info()[1]
                raise RuntimeError(
                    "Failed to set value for param=%s, index=%s, value=%s.\n"
                    "\tsource error message=%s"
                    % (self.name, str(key), str(val), str(msg))
                )
            #
            # Flag that things are fully constructed now (and changing an
            # immutable Param is now an exception).
            #
            self._constructed = True

            # populate all other indices with default data
            # (avoids calling _set_contains on self._index_set at runtime)
            if self._dense_initialize:
                self.to_dense_data()
        finally:
            timer.report()

    def _pprint(self):
        """
        Return data that will be printed for this component.
        """
        if self._default_val is Param.NoValue:
            default = "None"  # for backwards compatibility in reporting
        elif type(self._default_val) is types.FunctionType:
            default = "(function)"
        else:
            default = str(self._default_val)
        if self._mutable or not self.is_indexed():
            dataGen = lambda k, v: [v._value]
        else:
            dataGen = lambda k, v: [v]
        if self.index_set().isfinite() or self._default_val is Param.NoValue:
            _len = len(self)
        else:
            _len = 'inf'
        headers = [
            ("Size", _len),
            ("Index", self._index_set if self.is_indexed() else None),
            ("Domain", self.domain.name),
            ("Default", default),
            ("Mutable", self._mutable),
        ]
        if self._units is not None:
            headers.append(('Units', str(self._units)))
        return (headers, self.sparse_items, ("Value",), dataGen)


class ScalarParam(ParamData, Param):
    def __init__(self, *args, **kwds):
        ParamData.__init__(self, component=self)
        Param.__init__(self, *args, **kwds)
        self._index = UnindexedComponent_index

    #
    # Since this class derives from Component and Component.__getstate__
    # just packs up the entire __dict__ into the state dict, there s
    # nothing special that we need to do here.  We will just defer to the
    # super() get/set state.  Since all of our get/set state methods
    # rely on super() to traverse the MRO, this will automatically pick
    # up both the Component and Data base classes.
    #

    def __call__(self, exception=NOTSET):
        """
        Return the value of this parameter.
        """
        exception = _type_check_exception_arg(self, exception)

        if self._constructed:
            if not self._data:
                if self._mutable:
                    # This will trigger populating the _data dict and setting
                    # the _default, if applicable
                    self[None]
                else:
                    # Immutable Param defaults never get added to the
                    # _data dict
                    return self[None]
            return super(ScalarParam, self).__call__(exception=exception)
        if exception:
            raise ValueError(
                "Evaluating the numeric value of parameter '%s' before\n\t"
                "the Param has been constructed (there is currently no "
                "value to return)." % (self.name,)
            )

    def set_value(self, value, index=NOTSET):
        if index is NOTSET:
            index = None
        if self._constructed and not self._mutable:
            _raise_modifying_immutable_error(self, index)
        if not self._data:
            self._data[index] = self
        super(ScalarParam, self).set_value(value, index)

    def is_constant(self):
        """Determine if this ScalarParam is constant (and can be eliminated)

        Returns False if either unconstructed or mutable, as it must be kept
        in expressions (as it either doesn't have a value yet or the value
        can change later.
        """
        return self._constructed and not self._mutable


class SimpleParam(metaclass=RenamedClass):
    __renamed__new_class__ = ScalarParam
    __renamed__version__ = '6.0'


class IndexedParam(Param):
    # Because IndexedParam can use a non-standard data store (i.e., the
    # values in the _data dict may not be ComponentData objects), we
    # need to override the normal scheme for pre-allocating
    # ComponentData objects during deepcopy.
    def _create_objects_for_deepcopy(self, memo, component_list):
        if self.mutable:
            # Normal indexed object; leverage base implementation
            return super()._create_objects_for_deepcopy(memo, component_list)
        # This is immutable; only add the container (not the _data) to
        # the component_list.
        _new = self.__class__.__new__(self.__class__)
        _ans = memo.setdefault(id(self), _new)
        if _ans is _new:
            component_list.append((self, _new))
        return _ans

    # Because CP supports indirection [the ability to index objects by
    # another (inter) Var] for certain types (including Var), we will
    # catch the normal RuntimeError and return a (variable)
    # GetItemExpression.
    #
    # FIXME: We should integrate this logic into the base implementation
    # of `__getitem__()`, including the recognition / differentiation
    # between potentially variable GetItemExpression objects and
    # "constant" GetItemExpression objects.  That will need to wait for
    # the expression rework [JDS; Nov 22].
    def __getitem__(self, args) -> ParamData:
        try:
            return super().__getitem__(args)
        except:
            tmp = args if args.__class__ is tuple else (args,)
            if any(
                hasattr(arg, 'is_potentially_variable')
                and arg.is_potentially_variable()
                for arg in tmp
            ):
                return GetItemExpression((self,) + tmp)
            raise



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from math import fabs
from pyomo.contrib.gdpopt.solve_subproblem import solve_subproblem
from pyomo.contrib.gdpopt.util import fix_discrete_problem_solution_in_subproblem
from pyomo.core import value
from pyomo.opt import TerminationCondition as tc


class _OAAlgorithmMixIn:
    def _fix_discrete_soln_solve_subproblem_and_add_cuts(
        self, discrete_prob_util_block, subprob_util_block, config
    ):
        with fix_discrete_problem_solution_in_subproblem(
            discrete_prob_util_block, subprob_util_block, self, config
        ):
            nlp_termination = solve_subproblem(subprob_util_block, self, config)
            if nlp_termination in {tc.optimal, tc.feasible}:
                primal_improved = self._update_bounds_after_solve(
                    'subproblem',
                    primal=value(subprob_util_block.obj.expr),
                    logger=config.logger,
                )
                if primal_improved:
                    self.update_incumbent(subprob_util_block)
                self._add_cuts_to_discrete_problem(
                    subprob_util_block,
                    discrete_prob_util_block,
                    self.objective_sense,
                    config,
                    self.timing,
                )
            elif nlp_termination == tc.unbounded:
                # the whole problem is unbounded, we can stop
                self._update_primal_bound_to_unbounded(config)

        return nlp_termination not in {tc.infeasible, tc.unbounded}

    # Utility used in cut_generation: We saved a map of Disjuncts to the active
    # constraints they contain on the discrete problem util_block, and use it
    # here to find the active constraints under the current discrete
    # solution. Note that this preprocesses not just to be efficient, but
    # because everything on the Disjuncts is deactivated at this point, since
    # we've already transformed the discrete problem to a MILP
    def _get_active_untransformed_constraints(self, util_block, config):
        """Yield constraints in disjuncts where the indicator value is set or
        fixed to True."""
        model = util_block.parent_block()
        # Get active global constraints
        for constr in util_block.global_constraint_list:
            yield constr
        # get all the disjuncts in the original model. Check which ones are
        # True.
        for disj, constr_list in util_block.constraints_by_disjunct.items():
            if fabs(disj.binary_indicator_var.value - 1) <= config.integer_tolerance:
                for constr in constr_list:
                    yield constr



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for DataPortal objects
#

from itertools import zip_longest
import json
import os
from os.path import abspath, dirname, join

pyomo_dir = dirname(dirname(abspath(__file__))) + os.sep + ".."

import pyomo.common.unittest as unittest

from pyomo.common.errors import ApplicationError
from pyomo.common.tee import capture_output
from pyomo.dataportal.factory import DataManagerFactory
from pyomo.environ import (
    AbstractModel,
    ConcreteModel,
    Set,
    DataPortal,
    Param,
    Boolean,
    Any,
    value,
)

currdir = dirname(abspath(__file__)) + os.sep
example_dir = (
    pyomo_dir
    + os.sep
    + ".."
    + os.sep
    + "examples"
    + os.sep
    + "pyomo"
    + os.sep
    + "tutorials"
    + os.sep
    + "tab"
    + os.sep
)
tutorial_dir = (
    pyomo_dir
    + os.sep
    + ".."
    + os.sep
    + "examples"
    + os.sep
    + "pyomo"
    + os.sep
    + "tutorials"
    + os.sep
)

try:
    xls_interface = DataManagerFactory('xls').available()
except:
    xls_interface = False
try:
    xlsx_interface = DataManagerFactory('xlsx').available()
except:
    xlsx_interface = False
try:
    xlsb_interface = DataManagerFactory('xlsb').available()
except:
    xlsb_interface = False
try:
    xlsm_interface = DataManagerFactory('xlsm').available()
except:
    xlsm_interface = False
try:
    yaml_interface = DataManagerFactory('yaml').available()
    import yaml
except:
    yaml_interface = False


@unittest.skipIf(not xls_interface, "No XLS interface available")
class PyomoTableData(unittest.TestCase):
    def setUp(self):
        pass

    def construct(self, filename):
        pass

    def test_read_set(self):
        td = DataManagerFactory('xls')
        td.initialize(
            filename=currdir + "Book1.xls", range="TheRange", format='set', set="X"
        )
        try:
            td.open()
            td.read()
            td.close()
            self.assertEqual(
                td._info,
                [
                    'set',
                    'X',
                    ':=',
                    ('A1', 2.0, 3.0, 4.0),
                    ('A5', 6.0, 7.0, 8.0),
                    ('A9', 10.0, 11.0, 12.0),
                    ('A13', 14.0, 15.0, 16.0),
                ],
            )
        except ApplicationError:
            pass

    def test_read_param1(self):
        td = DataManagerFactory('xls')
        td.initialize(
            filename=currdir + "Book1.xls", range="TheRange", param=['bb', 'cc', 'dd']
        )
        try:
            td.open()
            td.read()
            td.close()
            self.assertEqual(
                td._info,
                [
                    'param',
                    ':',
                    'bb',
                    'cc',
                    'dd',
                    ':=',
                    'A1',
                    2.0,
                    3.0,
                    4.0,
                    'A5',
                    6.0,
                    7.0,
                    8.0,
                    'A9',
                    10.0,
                    11.0,
                    12.0,
                    'A13',
                    14.0,
                    15.0,
                    16.0,
                ],
            )
        except ApplicationError:
            pass

    def test_read_param2(self):
        td = DataManagerFactory('xls')
        td.initialize(
            filename=currdir + "Book1.xls",
            range="TheRange",
            index="X",
            param=['bb', 'cc', 'dd'],
        )
        try:
            td.open()
            td.read()
            td.close()
            self.assertEqual(
                td._info,
                [
                    'param',
                    ':',
                    'X',
                    ':',
                    'bb',
                    'cc',
                    'dd',
                    ':=',
                    'A1',
                    2.0,
                    3.0,
                    4.0,
                    'A5',
                    6.0,
                    7.0,
                    8.0,
                    'A9',
                    10.0,
                    11.0,
                    12.0,
                    'A13',
                    14.0,
                    15.0,
                    16.0,
                ],
            )
        except ApplicationError:
            pass

    def test_read_param3(self):
        td = DataManagerFactory('xls')
        td.initialize(
            filename=currdir + "Book1.xls", range="TheRange", index="X", param=["a"]
        )
        try:
            td.open()
            td.read()
            td.close()
            self.assertEqual(
                td._info,
                [
                    'param',
                    ':',
                    'X',
                    ':',
                    'a',
                    ':=',
                    'A1',
                    2.0,
                    3.0,
                    4.0,
                    'A5',
                    6.0,
                    7.0,
                    8.0,
                    'A9',
                    10.0,
                    11.0,
                    12.0,
                    'A13',
                    14.0,
                    15.0,
                    16.0,
                ],
            )
        except ApplicationError:
            pass

    def test_read_param4(self):
        td = DataManagerFactory('xls')
        td.initialize(
            filename=currdir + "Book1.xls",
            range="TheRange",
            index="X",
            param=['a', 'b'],
        )
        try:
            td.open()
            td.read()
            td.close()
            self.assertEqual(
                td._info,
                [
                    'param',
                    ':',
                    'X',
                    ':',
                    'a',
                    'b',
                    ':=',
                    'A1',
                    2.0,
                    3.0,
                    4.0,
                    'A5',
                    6.0,
                    7.0,
                    8.0,
                    'A9',
                    10.0,
                    11.0,
                    12.0,
                    'A13',
                    14.0,
                    15.0,
                    16.0,
                ],
            )
        except ApplicationError:
            pass

    def test_read_array1(self):
        td = DataManagerFactory('xls')
        td.initialize(
            filename=currdir + "Book1.xls", range="TheRange", param="X", format="array"
        )
        try:
            td.open()
            td.read()
            td.close()
            self.assertEqual(
                td._info,
                [
                    'param',
                    'X',
                    ':',
                    'bb',
                    'cc',
                    'dd',
                    ':=',
                    'A1',
                    2.0,
                    3.0,
                    4.0,
                    'A5',
                    6.0,
                    7.0,
                    8.0,
                    'A9',
                    10.0,
                    11.0,
                    12.0,
                    'A13',
                    14.0,
                    15.0,
                    16.0,
                ],
            )
        except ApplicationError:
            pass

    def test_read_array2(self):
        td = DataManagerFactory('xls')
        td.initialize(
            filename=currdir + "Book1.xls",
            range="TheRange",
            param="X",
            format="transposed_array",
        )
        try:
            td.open()
            td.read()
            td.close()
            self.assertEqual(
                td._info,
                [
                    'param',
                    'X',
                    '(tr)',
                    ':',
                    'bb',
                    'cc',
                    'dd',
                    ':=',
                    'A1',
                    2.0,
                    3.0,
                    4.0,
                    'A5',
                    6.0,
                    7.0,
                    8.0,
                    'A9',
                    10.0,
                    11.0,
                    12.0,
                    'A13',
                    14.0,
                    15.0,
                    16.0,
                ],
            )
        except ApplicationError:
            pass

    def test_error1(self):
        td = DataManagerFactory('xls')
        td.initialize(filename="bad")
        try:
            td.open()
            self.fail("Expected IOError because of bad file")
        except IOError:
            pass

    def test_error2(self):
        td = DataManagerFactory('xls')
        try:
            td.open()
            self.fail("Expected IOError because no file specified")
        except (IOError, AttributeError):
            pass

    def test_error3(self):
        td = DataManagerFactory('txt')
        try:
            td.initialize(filename=currdir + "display.txt")
            td.open()
            self.fail("Expected IOError because of bad file type")
        except (IOError, AttributeError):
            pass

    def test_error4(self):
        td = DataManagerFactory('txt')
        try:
            td.initialize(filename=currdir + "dummy")
            td.open()
            self.fail("Expected IOError because of bad file type")
        except (IOError, AttributeError):
            pass

    def test_error5(self):
        td = DataManagerFactory('tab')
        td.initialize(filename=example_dir + "D.tab", param="D", format="foo")
        td.open()
        try:
            td.read()
            self.fail("Expected IOError because of bad format")
        except ValueError:
            pass


class PyomoDataPortal(unittest.TestCase):
    def test_tableA1_1(self):
        # Importing a single column of data
        model = AbstractModel()
        model.A = Set()
        data = DataPortal(filename=os.path.abspath(example_dir + 'A.tab'), set=model.A)
        self.assertEqual(set(data['A']), set(['A1', 'A2', 'A3']))
        instance = model.create_instance(data)
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3']))

    def test_tableA1_2(self):
        # Importing a single column of data
        model = AbstractModel()
        model.A = Set()
        data = DataPortal()
        data.load(filename=os.path.abspath(example_dir + 'A.tab'), set=model.A)
        instance = model.create_instance(data)
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3']))

    def test_tableA1_3(self):
        # Importing a single column of data
        model = AbstractModel()
        model.A = Set()
        data = DataPortal()
        data.connect(filename=os.path.abspath(example_dir + 'B.tab'))
        # The first connection will be closed here
        data.connect(filename=os.path.abspath(example_dir + 'A.tab'))
        data.load(set=model.A)
        data.disconnect()
        instance = model.create_instance(data)
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3']))

    def test_md1(self):
        md = DataPortal()
        md.connect(filename=example_dir + "A.tab")
        try:
            md.load()
            self.fail("Must specify a model")
        except ValueError:
            pass
        model = AbstractModel()
        try:
            md.load(model=model)
            self.fail("Expected ValueError")
        except ValueError:
            pass
        model.A = Set()

    def test_md2(self):
        md = DataPortal()
        model = AbstractModel()
        model.A = Set()
        md.load(model=model, filename=currdir + "data1.dat")
        self.assertEqual(set(md['A']), set([1, 2, 3]))

    def test_md3(self):
        md = DataPortal()
        model = AbstractModel()
        model.A = Set()
        try:
            md.load(model=model, filename=currdir + "data2.dat")
            self.fail("Expected error because of extraneous text")
        except IOError:
            pass

    def test_md4(self):
        md = DataPortal()
        model = AbstractModel()
        model.A = Set()
        model.B = Set()
        model.C = Set()
        md.load(model=model, filename=currdir + "data3.dat")
        self.assertEqual(set(md['A']), set([]))
        self.assertEqual(set(md['B']), set([(1, 2)]))
        self.assertEqual(set(md['C']), set([('a', 'b', 'c')]))

    def test_md5(self):
        md = DataPortal()
        model = AbstractModel()
        model.A = Set()
        try:
            md.load(model=model, filename=currdir + "data4.dat")
        except (ValueError, IOError):
            pass

    def test_md6(self):
        md = DataPortal()
        model = AbstractModel()
        model.A = Set()
        try:
            md.load(model=model, filename=currdir + "data5.dat")
        except ValueError:
            pass

    def test_md7(self):
        md = DataPortal()
        model = AbstractModel()
        try:
            md.load(model=model, filename=currdir + "data1.tab")
            self.fail("Expected IOError")
        except IOError:
            pass

    def test_md8(self):
        md = DataPortal()
        model = AbstractModel()
        model.A = Set()
        try:
            md.load(model=model, filename=currdir + "data6.dat")
            self.fail("Expected IOError")
        except IOError:
            pass

    def test_md9(self):
        md = DataPortal()
        model = AbstractModel()
        model.A = Set()
        model.B = Param(model.A)
        md.load(model=model, filename=currdir + "data7.dat")
        self.assertEqual(set(md['A']), set(['a', 'b', 'c']))
        self.assertEqual(md['B'], {'a': 1.0, 'c': 3.0})

    def test_md10(self):
        md = DataPortal()
        model = AbstractModel()
        model.A = Param(within=Boolean)
        model.B = Param(within=Boolean)
        model.Z = Set()
        model.Y = Set(model.Z)
        md.load(model=model, filename=currdir + "data8.dat")
        self.assertEqual(md['A'], False)
        self.assertEqual(md['B'], True)
        self.assertEqual(
            md['Z'], ['foo[*]', 'bar[ * ]', 'bar[1,*,a,*]', 'foo-bar', 'hello-goodbye']
        )
        self.assertEqual(
            md['Y']['foo-bar'],
            ['foo[*]', 'bar[ * ]', 'bar[1,*,a,*]', 'foo-bar', 'hello-goodbye'],
        )
        instance = model.create_instance(md)

    def test_md11(self):
        cwd = os.getcwd()
        os.chdir(currdir)
        md = DataPortal()
        model = AbstractModel()
        model.A = Set()
        model.B = Set()
        model.C = Set()
        model.D = Set()
        md.load(model=model, filename=currdir + "data11.dat")
        self.assertEqual(set(md['A']), set([]))
        self.assertEqual(set(md['B']), set([(1, 2)]))
        self.assertEqual(set(md['C']), set([('a', 'b', 'c')]))
        self.assertEqual(set(md['D']), set([1, 3, 5]))
        os.chdir(cwd)

    def test_md11a(self):
        cwd = os.getcwd()
        os.chdir(currdir)
        model = AbstractModel()
        model.a = Param()
        model.b = Param()
        model.c = Param()
        model.d = Param()
        # Test 1
        instance = model.create_instance(
            currdir + 'data14.dat', namespaces=['ns1', 'ns2']
        )
        self.assertEqual(value(instance.a), 1)
        self.assertEqual(value(instance.b), 2)
        self.assertEqual(value(instance.c), 2)
        self.assertEqual(value(instance.d), 2)
        # Test 2
        instance = model.create_instance(
            currdir + 'data14.dat', namespaces=['ns1', 'ns3', 'nsX']
        )
        self.assertEqual(value(instance.a), 1)
        self.assertEqual(value(instance.b), 100)
        self.assertEqual(value(instance.c), 3)
        self.assertEqual(value(instance.d), 100)
        # Test None
        instance = model.create_instance(currdir + 'data14.dat')
        self.assertEqual(value(instance.a), -1)
        self.assertEqual(value(instance.b), -2)
        self.assertEqual(value(instance.c), -3)
        self.assertEqual(value(instance.d), -4)
        #
        os.chdir(cwd)

    def test_md12(self):
        model = ConcreteModel()
        model.A = Set()
        md = DataPortal()
        try:
            md.load(filename=example_dir + 'A.tab', format='bad', set=model.A)
            self.fail("Bad format error")
        except ValueError:
            pass
        try:
            md.load(filename=example_dir + 'A.tab')
            self.fail("Bad format error")
        except ValueError:
            pass

    @unittest.expectedFailure
    def test_md13(self):
        md = DataPortal()
        model = AbstractModel()
        model.p = Param()
        instance = model.create_instance(currdir + "data15.dat")
        md.load(model=model, filename=currdir + "data15.dat")
        try:
            md.load(model=model, filename=currdir + "data15.dat")
            self.fail("Expected IOError")
        except IOError:
            pass

    def test_md14(self):
        try:
            md = DataPortal(1)
            self.fail("Expected RuntimeError")
        except RuntimeError:
            pass
        try:
            md = DataPortal(foo=True)
            self.fail("Expected ValueError")
        except ValueError:
            pass

    def test_md15(self):
        md = DataPortal()
        try:
            md.connect(filename='foo.dummy')
            self.fail("Expected OSError")
        except IOError:
            pass
        except OSError:
            pass

    def test_md16(self):
        md = DataPortal()
        try:
            md.data(namespace='foo')
            self.fail("Expected IOError")
        except IOError:
            pass

    def test_md17(self):
        md = DataPortal()
        try:
            md[1, 2, 3, 4]
            self.fail("Expected IOError")
        except IOError:
            pass

    def test_md18(self):
        cwd = os.getcwd()
        os.chdir(currdir)

        md = DataPortal()
        md.load(filename=currdir + "data17.dat")

        self.assertEqual(md['A'], 1)
        self.assertEqual(md['B'], {'a': 1})
        self.assertEqual(md['C'], {'a': 1, 'b': 2, 'c': 3})
        self.assertEqual(md['D'], 1)

        os.chdir(cwd)

    def test_dat_type_conversion(self):
        model = AbstractModel()
        model.I = Set()
        model.p = Param(model.I, domain=Any)
        i = model.create_instance(currdir + "data_types.dat")
        ref = {
            50: (int, 2),
            55: (int, -2),
            51: (int, 200),
            52: (int, -200),
            53: (float, 0.02),
            54: (float, -0.02),
            10: (float, 1.0),
            11: (float, -1.0),
            12: (float, 0.1),
            13: (float, -0.1),
            14: (float, 1.1),
            15: (float, -1.1),
            20: (float, 200.0),
            21: (float, -200.0),
            22: (float, 0.02),
            23: (float, -0.02),
            30: (float, 210.0),
            31: (float, -210.0),
            32: (float, 0.021),
            33: (float, -0.021),
            40: (float, 10.0),
            41: (float, -10.0),
            42: (float, 0.001),
            43: (float, -0.001),
            1000: (str, "a_string"),
            1001: (str, "a_string"),
            1002: (str, 'a_string'),
            1003: (str, 'a " string'),
            1004: (str, "a ' string"),
            1005: (str, '1234_567'),
            1006: (str, '123'),
        }
        for k, v in i.p.items():
            # print(k,v, type(v))
            if k in ref:
                err = "index %s: (%s, %s) does not match ref %s" % (
                    k,
                    type(v),
                    v,
                    ref[k],
                )
                self.assertIs(type(v), ref[k][0], err)
                self.assertEqual(v, ref[k][1], err)
            else:
                n = k // 10
                err = "index %s: (%s, %s) does not match ref %s" % (
                    k,
                    type(v),
                    v,
                    ref[n],
                )
                self.assertIs(type(v), ref[n][0], err)
                self.assertEqual(v, ref[n][1], err)

    def test_data_namespace(self):
        model = AbstractModel()
        model.a = Param()
        model.b = Param()
        model.c = Param()
        model.d = Param()
        model.A = Set()
        model.e = Param(model.A)
        md = DataPortal()
        md.load(model=model, filename=currdir + "data16.dat")
        # data()
        self.assertEqual(
            md.data(namespace='ns1'),
            {'a': {None: 1}, 'A': {None: [7, 9, 11]}, 'e': {9: 90, 7: 70, 11: 110}},
        )
        # __getitem__
        self.assertEqual(md['ns1', 'a'], 1)
        # namespaces()
        self.assertEqual(
            sorted(md.namespaces(), key=lambda x: 'None' if x is None else x),
            [None, 'ns1', 'ns2', 'ns3', 'nsX'],
        )
        # keys()
        self.assertEqual(sorted(md.keys()), ['A', 'a', 'b', 'c', 'd', 'e'])
        self.assertEqual(sorted(md.keys('ns1')), ['A', 'a', 'e'])
        # values()
        self.assertEqual(
            sorted(
                md.values(),
                key=lambda x: (
                    tuple(sorted(x) + [0])
                    if type(x) is list
                    else tuple(sorted(x.values())) if not type(x) is int else (x,)
                ),
            ),
            [-4, -3, -2, -1, [1, 3, 5], {1: 10, 3: 30, 5: 50}],
        )
        self.assertEqual(
            sorted(
                md.values('ns1'),
                key=lambda x: (
                    tuple(sorted(x) + [0])
                    if type(x) is list
                    else tuple(sorted(x.values())) if not type(x) is int else (x,)
                ),
            ),
            [1, [7, 9, 11], {7: 70, 9: 90, 11: 110}],
        )
        # items()
        self.assertEqual(
            sorted(md.items()),
            [
                ('A', [1, 3, 5]),
                ('a', -1),
                ('b', -2),
                ('c', -3),
                ('d', -4),
                ('e', {1: 10, 3: 30, 5: 50}),
            ],
        )
        self.assertEqual(
            sorted(md.items('ns1')),
            [('A', [7, 9, 11]), ('a', 1), ('e', {7: 70, 9: 90, 11: 110})],
        )


class TestOnlyTextPortal(unittest.TestCase):
    suffix = '.tab'
    skiplist = []

    def check_skiplist(self, name):
        if name in self.skiplist:
            self.skipTest('Skipping test %s' % name)

    def create_options(self, name):
        return {
            'filename': os.path.abspath(
                tutorial_dir + os.sep + 'tab' + os.sep + name + self.suffix
            )
        }

    def create_write_options(self, name):
        return {
            'filename': os.path.abspath(currdir + os.sep + name + self.suffix),
            'sort': True,
        }

    def test_empty(self):
        # Importing an empty file
        self.check_skiplist('empty')
        dp = DataPortal()
        try:
            dp.load(
                set='A',
                filename=os.path.abspath(currdir + os.sep + 'empty' + self.suffix),
            )
            self.fail("Expected IOError")
        except IOError:
            pass

    def test_tableA(self):
        # Importing an unordered set of arbitrary data
        self.check_skiplist('tableA')
        dp = DataPortal()
        dp.load(set='A', **self.create_options('A'))
        self.assertEqual(set(dp.data('A')), set(['A1', 'A2', 'A3']))

    def test_tableB(self):
        # Importing an unordered set of numeric data
        self.check_skiplist('tableB')
        dp = DataPortal()
        dp.load(set='B', **self.create_options('B'))
        self.assertEqual(set(dp.data('B')), set([1, 2, 3]))

    def test_tableC(self):
        # Importing a multi-column table, where all columns are
        # treated as values for a set with tuple values.
        self.check_skiplist('tableC')
        dp = DataPortal()
        dp.load(set='C', **self.create_options('C'))
        self.assertEqual(
            set(dp.data('C')),
            set(
                [
                    ('A1', 1),
                    ('A1', 2),
                    ('A1', 3),
                    ('A2', 1),
                    ('A2', 2),
                    ('A2', 3),
                    ('A3', 1),
                    ('A3', 2),
                    ('A3', 3),
                ]
            ),
        )

    def test_tableD(self):
        # Importing a 2D array of data as a set.
        self.check_skiplist('tableD')
        dp = DataPortal()
        dp.load(set='D', format='set_array', **self.create_options('D'))
        self.assertEqual(set(dp.data('D')), set([('A1', 1), ('A2', 2), ('A3', 3)]))

    def test_tableZ(self):
        # Importing a single parameter
        self.check_skiplist('tableZ')
        dp = DataPortal()
        dp.load(param='Z', **self.create_options('Z'))
        self.assertEqual(dp.data('Z'), 1.01)

    def test_tableY(self):
        # Same as tableXW.
        self.check_skiplist('tableY')
        dp = DataPortal()
        dp.load(param='Y', **self.create_options('Y'))
        self.assertEqual(dp.data('Y'), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})

    def test_tableXW_1(self):
        # Importing a table, but only reporting the values for the non-index
        # parameter columns.  The first column is assumed to represent an
        # index column.
        self.check_skiplist('tableXW_1')
        dp = DataPortal()
        dp.load(param=('X', 'W'), **self.create_options('XW'))
        self.assertEqual(dp.data('X'), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(dp.data('W'), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})

    def test_tableXW_3(self):
        # Like test_tableXW_1, except that set A is defined in the load statement.
        self.check_skiplist('tableXW_3')
        dp = DataPortal()
        dp.load(index='A', param=('X', 'W'), **self.create_options('XW'))
        self.assertEqual(set(dp.data('A')), set(['A1', 'A2', 'A3']))
        self.assertEqual(dp.data('X'), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(dp.data('W'), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})

    def test_tableXW_4(self):
        # Like test_tableXW_1, except that set A is defined in the load statement and all values are mapped.
        self.check_skiplist('tableXW_4')
        dp = DataPortal()
        dp.load(
            select=('A', 'W', 'X'),
            index='B',
            param=('R', 'S'),
            **self.create_options('XW'),
        )
        self.assertEqual(set(dp.data('B')), set(['A1', 'A2', 'A3']))
        self.assertEqual(dp.data('S'), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(dp.data('R'), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})

    def test_tableT(self):
        # Importing a 2D array of parameters that are transposed.
        self.check_skiplist('tableT')
        dp = DataPortal()
        dp.load(format='transposed_array', param='T', **self.create_options('T'))
        self.assertEqual(
            dp.data('T'),
            {
                ('A2', 'I1'): 2.3,
                ('A1', 'I2'): 1.4,
                ('A1', 'I3'): 1.5,
                ('A1', 'I4'): 1.6,
                ('A1', 'I1'): 1.3,
                ('A3', 'I4'): 3.6,
                ('A2', 'I4'): 2.6,
                ('A3', 'I1'): 3.3,
                ('A2', 'I3'): 2.5,
                ('A3', 'I2'): 3.4,
                ('A2', 'I2'): 2.4,
                ('A3', 'I3'): 3.5,
            },
        )

    def test_tableU(self):
        # Importing a 2D array of parameters.
        self.check_skiplist('tableU')
        dp = DataPortal()
        dp.load(format='array', param='U', **self.create_options('U'))
        self.assertEqual(
            dp.data('U'),
            {
                ('I2', 'A1'): 1.4,
                ('I3', 'A1'): 1.5,
                ('I3', 'A2'): 2.5,
                ('I4', 'A1'): 1.6,
                ('I3', 'A3'): 3.5,
                ('I1', 'A2'): 2.3,
                ('I4', 'A3'): 3.6,
                ('I1', 'A3'): 3.3,
                ('I4', 'A2'): 2.6,
                ('I2', 'A3'): 3.4,
                ('I1', 'A1'): 1.3,
                ('I2', 'A2'): 2.4,
            },
        )

    def test_tableS(self):
        # Importing a table, but only reporting the values for the non-index
        # parameter columns.  The first column is assumed to represent an
        # index column.  A missing value is represented in the column data.
        self.check_skiplist('tableS')
        dp = DataPortal()
        dp.load(param='S', **self.create_options('S'))
        self.assertEqual(dp.data('S'), {'A1': 3.3, 'A3': 3.5})

    def test_tablePO(self):
        # Importing a table that has multiple indexing columns
        self.check_skiplist('tablePO')
        dp = DataPortal()
        dp.load(index='J', param=('P', 'O'), **self.create_options('PO'))
        self.assertEqual(
            set(dp.data('J')), set([('A3', 'B3'), ('A1', 'B1'), ('A2', 'B2')])
        )
        self.assertEqual(
            dp.data('P'), {('A3', 'B3'): 4.5, ('A1', 'B1'): 4.3, ('A2', 'B2'): 4.4}
        )
        self.assertEqual(
            dp.data('O'), {('A3', 'B3'): 5.5, ('A1', 'B1'): 5.3, ('A2', 'B2'): 5.4}
        )

    def test_tablePP(self):
        # Importing a table that has a 2-d indexing
        self.check_skiplist('tablePP')
        dp = DataPortal()
        dp.load(param='PP', **self.create_options('PP'))
        # self.assertEqual(set(dp.data('J')), set([('A3', 'B3'), ('A1', 'B1'), ('A2', 'B2')]) )
        self.assertEqual(
            dp.data('PP'), {('A3', 'B3'): 4.5, ('A1', 'B1'): 4.3, ('A2', 'B2'): 4.4}
        )


class TestOnlyCsvPortal(TestOnlyTextPortal):
    suffix = '.csv'

    def create_options(self, name):
        return {
            'filename': os.path.abspath(
                tutorial_dir + os.sep + 'csv' + os.sep + name + self.suffix
            )
        }


class TestOnlyXmlPortal(TestOnlyTextPortal):
    suffix = '.xml'
    skiplist = ['tableD', 'tableT', 'tableU']

    def create_options(self, name):
        return {
            'filename': os.path.abspath(
                tutorial_dir + os.sep + 'xml' + os.sep + name + self.suffix
            )
        }


class TestOnlyJsonPortal(TestOnlyTextPortal):
    suffix = '.json'
    skiplist = ['tableD', 'tableT', 'tableU', 'tableXW_4']

    def create_options(self, name):
        return {
            'filename': os.path.abspath(
                tutorial_dir + os.sep + 'json' + os.sep + name + self.suffix
            )
        }

    def compare_data(self, name, file_suffix):
        if file_suffix == '.json':
            with (
                open(join(currdir, name + file_suffix), 'r') as out,
                open(join(currdir, name + '.baseline' + file_suffix), 'r') as txt,
            ):
                self.assertStructuredAlmostEqual(
                    json.load(txt), json.load(out), allow_second_superset=True, abstol=0
                )
        elif file_suffix == '.yaml':
            with (
                open(join(currdir, name + file_suffix), 'r') as out,
                open(join(currdir, name + '.baseline' + file_suffix), 'r') as txt,
            ):
                self.assertStructuredAlmostEqual(
                    yaml.full_load(txt),
                    yaml.full_load(out),
                    allow_second_superset=True,
                    abstol=0,
                )
        else:
            with (
                open(join(currdir, name + file_suffix), 'r') as f1,
                open(join(currdir, name + '.baseline' + file_suffix), 'r') as f2,
            ):
                f1_contents = list(filter(None, f1.read().split()))
                f2_contents = list(filter(None, f2.read().split()))
                for item1, item2 in zip_longest(f1_contents, f2_contents):
                    self.assertEqual(item1, item2)
        os.remove(currdir + name + file_suffix)

    def test_store_set1(self):
        # Write 1-D set
        model = ConcreteModel()
        model.A = Set(initialize=[1, 3, 5])
        data = DataPortal()
        data.store(data=model.A, **self.create_write_options('set1'))
        self.compare_data('set1', self.suffix)

    def test_store_set1a(self):
        # Write 1-D set
        model = ConcreteModel()
        model.A = Set(initialize=[1, 3, 5])
        data = DataPortal()
        data.store(data="A", model=model, **self.create_write_options('set1'))
        self.compare_data('set1', self.suffix)

    def test_store_set2(self):
        # Write 2-D set
        model = ConcreteModel()
        model.A = Set(initialize=[(1, 2), (3, 4), (5, 6)], dimen=2)
        data = DataPortal()
        data.store(data=model.A, **self.create_write_options('set2'))
        self.compare_data('set2', self.suffix)

    def test_store_param1(self):
        # Write scalar param
        model = ConcreteModel()
        model.p = Param(initialize=1)
        data = DataPortal()
        data.store(data=model.p, **self.create_write_options('param1'))
        self.compare_data('param1', self.suffix)

    def test_store_param2(self):
        # Write 1-D param
        model = ConcreteModel()
        model.A = Set(initialize=[1, 2, 3])
        model.p = Param(model.A, initialize={1: 10, 2: 20, 3: 30})
        data = DataPortal()
        data.store(data=model.p, **self.create_write_options('param2'))
        self.compare_data('param2', self.suffix)

    def test_store_param3(self):
        # Write 2-D params
        model = ConcreteModel()
        model.A = Set(initialize=[(1, 2), (2, 3), (3, 4)], dimen=2)
        model.p = Param(model.A, initialize={(1, 2): 10, (2, 3): 20, (3, 4): 30})
        model.q = Param(model.A, initialize={(1, 2): 11, (2, 3): 21, (3, 4): 31})
        data = DataPortal()
        data.store(data=(model.p, model.q), **self.create_write_options('param3'))
        self.compare_data('param3', self.suffix)

    def test_store_param4(self):
        # Write 2-D params
        model = ConcreteModel()
        model.A = Set(initialize=[(1, 2), (2, 3), (3, 4)], dimen=2)
        model.p = Param(model.A, initialize={(1, 2): 10, (2, 3): 20, (3, 4): 30})
        model.q = Param(model.A, initialize={(1, 2): 11, (2, 3): 21, (3, 4): 31})
        data = DataPortal()
        data.store(
            data=(model.p, model.q),
            columns=('a', 'b', 'c', 'd'),
            **self.create_write_options('param4'),
        )
        self.compare_data('param4', self.suffix)


@unittest.skipIf(not yaml_interface, "No YAML interface available")
class TestOnlyYamlPortal(TestOnlyJsonPortal):
    suffix = '.yaml'

    def create_options(self, name):
        return {
            'filename': os.path.abspath(
                tutorial_dir + os.sep + 'yaml' + os.sep + name + self.suffix
            )
        }


class TestTextPortal(unittest.TestCase):
    suffix = '.tab'
    skiplist = []

    def check_skiplist(self, name):
        if name in self.skiplist:
            self.skipTest('Skipping test %s' % name)

    def create_options(self, name):
        return {
            'filename': os.path.abspath(
                tutorial_dir + os.sep + 'tab' + os.sep + name + self.suffix
            )
        }

    def create_write_options(self, name):
        return {
            'filename': os.path.abspath(currdir + os.sep + name + self.suffix),
            'sort': True,
        }

    def compare_data(self, name, file_suffix):
        if file_suffix == '.json':
            with (
                open(join(currdir, name + file_suffix), 'r') as out,
                open(join(currdir, name + '.baseline' + file_suffix), 'r') as txt,
            ):
                self.assertStructuredAlmostEqual(
                    json.load(txt), json.load(out), allow_second_superset=True, abstol=0
                )
        elif file_suffix == '.yaml':
            with (
                open(join(currdir, name + file_suffix), 'r') as out,
                open(join(currdir, name + '.baseline' + file_suffix), 'r') as txt,
            ):
                self.assertStructuredAlmostEqual(
                    yaml.full_load(txt),
                    yaml.full_load(out),
                    allow_second_superset=True,
                    abstol=0,
                )
        else:
            try:
                with (
                    open(join(currdir, name + file_suffix), 'r') as f1,
                    open(join(currdir, name + '.baseline' + file_suffix), 'r') as f2,
                ):
                    f1_contents = list(filter(None, f1.read().split()))
                    f2_contents = list(filter(None, f2.read().split()))
                    for item1, item2 in zip_longest(f1_contents, f2_contents):
                        self.assertEqual(item1, item2)
            except:
                with (
                    open(join(currdir, name + file_suffix), 'r') as out,
                    open(join(currdir, name + '.baseline' + file_suffix), 'r') as txt,
                ):
                    self.assertEqual(
                        out.read().strip().replace(' ', ''),
                        txt.read().strip().replace(' ', ''),
                    )
        os.remove(currdir + name + file_suffix)

    def test_tableA(self):
        # Importing an unordered set of arbitrary data
        self.check_skiplist('tableA')
        model = AbstractModel()
        model.A = Set()
        data = DataPortal()
        data.load(set=model.A, **self.create_options('A'))
        instance = model.create_instance(data)
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3']))

    def test_tableB(self):
        # Importing an unordered set of numeric data
        self.check_skiplist('tableB')
        model = AbstractModel()
        model.B = Set()
        data = DataPortal()
        data.load(set=model.B, **self.create_options('B'))
        instance = model.create_instance(data)
        self.assertEqual(set(instance.B.data()), set([1, 2, 3]))

    def test_tableC(self):
        # Importing a multi-column table, where all columns are
        # treated as values for a set with tuple values.
        self.check_skiplist('tableC')
        model = AbstractModel()
        model.C = Set(dimen=2)
        data = DataPortal()
        data.load(set=model.C, **self.create_options('C'))
        instance = model.create_instance(data)
        self.assertEqual(
            set(instance.C.data()),
            set(
                [
                    ('A1', 1),
                    ('A1', 2),
                    ('A1', 3),
                    ('A2', 1),
                    ('A2', 2),
                    ('A2', 3),
                    ('A3', 1),
                    ('A3', 2),
                    ('A3', 3),
                ]
            ),
        )

    def test_tableD(self):
        # Importing a 2D array of data as a set.
        self.check_skiplist('tableD')
        model = AbstractModel()
        model.C = Set(dimen=2)
        data = DataPortal()
        data.load(set=model.C, format='set_array', **self.create_options('D'))
        instance = model.create_instance(data)
        self.assertEqual(set(instance.C.data()), set([('A1', 1), ('A2', 2), ('A3', 3)]))

    def test_tableZ(self):
        # Importing a single parameter
        self.check_skiplist('tableZ')
        model = AbstractModel()
        model.Z = Param(default=99.0)
        data = DataPortal()
        data.load(param=model.Z, **self.create_options('Z'))
        instance = model.create_instance(data)
        self.assertEqual(instance.Z, 1.01)

    def test_tableY(self):
        # Same as tableXW.
        self.check_skiplist('tableY')
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3', 'A4'])
        model.Y = Param(model.A)
        data = DataPortal()
        data.load(param=model.Y, **self.create_options('Y'))
        instance = model.create_instance(data)
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3', 'A4']))
        self.assertEqual(instance.Y.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})

    def test_tableXW_1(self):
        # Importing a table, but only reporting the values for the non-index
        # parameter columns.  The first column is assumed to represent an
        # index column.
        self.check_skiplist('tableXW_1')
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3', 'A4'])
        model.X = Param(model.A)
        model.W = Param(model.A)
        data = DataPortal()
        data.load(param=(model.X, model.W), **self.create_options('XW'))
        instance = model.create_instance(data)
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3', 'A4']))
        self.assertEqual(instance.X.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(instance.W.extract_values(), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})

    def test_tableXW_2(self):
        # Like test_tableXW_1, except that set A is not defined.
        self.check_skiplist('tableXW_2')
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3'])
        model.X = Param(model.A)
        model.W = Param(model.A)
        data = DataPortal()
        data.load(param=(model.X, model.W), **self.create_options('XW'))
        instance = model.create_instance(data)
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3']))
        self.assertEqual(instance.X.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(instance.W.extract_values(), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})

    def test_tableXW_3(self):
        # Like test_tableXW_1, except that set A is defined in the load statement.
        self.check_skiplist('tableXW_3')
        model = AbstractModel()
        model.A = Set()
        model.X = Param(model.A)
        model.W = Param(model.A)
        data = DataPortal()
        data.load(index=model.A, param=(model.X, model.W), **self.create_options('XW'))
        instance = model.create_instance(data)
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3']))
        self.assertEqual(instance.X.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(instance.W.extract_values(), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})

    def test_tableXW_4(self):
        # Like test_tableXW_1, except that set A is defined in the load statement and all values are mapped.
        self.check_skiplist('tableXW_4')
        model = AbstractModel()
        model.B = Set()
        model.R = Param(model.B)
        model.S = Param(model.B)
        data = DataPortal()
        data.load(
            select=('A', 'W', 'X'),
            index=model.B,
            param=(model.R, model.S),
            **self.create_options('XW'),
        )
        instance = model.create_instance(data)
        self.assertEqual(set(instance.B.data()), set(['A1', 'A2', 'A3']))
        self.assertEqual(instance.S.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(instance.R.extract_values(), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})

    def test_tableT(self):
        # Importing a 2D array of parameters that are transposed.
        self.check_skiplist('tableT')
        model = AbstractModel()
        model.B = Set(initialize=['I1', 'I2', 'I3', 'I4'])
        model.A = Set(initialize=['A1', 'A2', 'A3'])
        model.T = Param(model.A, model.B)
        data = DataPortal()
        data.load(format='transposed_array', param=model.T, **self.create_options('T'))
        instance = model.create_instance(data)
        self.assertEqual(
            instance.T.extract_values(),
            {
                ('A2', 'I1'): 2.3,
                ('A1', 'I2'): 1.4,
                ('A1', 'I3'): 1.5,
                ('A1', 'I4'): 1.6,
                ('A1', 'I1'): 1.3,
                ('A3', 'I4'): 3.6,
                ('A2', 'I4'): 2.6,
                ('A3', 'I1'): 3.3,
                ('A2', 'I3'): 2.5,
                ('A3', 'I2'): 3.4,
                ('A2', 'I2'): 2.4,
                ('A3', 'I3'): 3.5,
            },
        )

    def test_tableU(self):
        # Importing a 2D array of parameters.
        self.check_skiplist('tableU')
        model = AbstractModel()
        model.A = Set(initialize=['I1', 'I2', 'I3', 'I4'])
        model.B = Set(initialize=['A1', 'A2', 'A3'])
        model.U = Param(model.A, model.B)
        data = DataPortal()
        data.load(format='array', param=model.U, **self.create_options('U'))
        instance = model.create_instance(data)
        self.assertEqual(
            instance.U.extract_values(),
            {
                ('I2', 'A1'): 1.4,
                ('I3', 'A1'): 1.5,
                ('I3', 'A2'): 2.5,
                ('I4', 'A1'): 1.6,
                ('I3', 'A3'): 3.5,
                ('I1', 'A2'): 2.3,
                ('I4', 'A3'): 3.6,
                ('I1', 'A3'): 3.3,
                ('I4', 'A2'): 2.6,
                ('I2', 'A3'): 3.4,
                ('I1', 'A1'): 1.3,
                ('I2', 'A2'): 2.4,
            },
        )

    def test_tableS(self):
        # Importing a table, but only reporting the values for the non-index
        # parameter columns.  The first column is assumed to represent an
        # index column.  A missing value is represented in the column data.
        self.check_skiplist('tableS')
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3', 'A4'])
        model.S = Param(model.A)
        data = DataPortal()
        data.load(param=model.S, **self.create_options('S'))
        instance = model.create_instance(data)
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3', 'A4']))
        self.assertEqual(instance.S.extract_values(), {'A1': 3.3, 'A3': 3.5})

    def test_tablePO(self):
        # Importing a table that has multiple indexing columns
        self.check_skiplist('tablePO')
        model = AbstractModel()
        model.J = Set(dimen=2)
        model.P = Param(model.J)
        model.O = Param(model.J)
        data = DataPortal()
        data.load(index=model.J, param=(model.P, model.O), **self.create_options('PO'))
        instance = model.create_instance(data)
        self.assertEqual(
            set(instance.J.data()), set([('A3', 'B3'), ('A1', 'B1'), ('A2', 'B2')])
        )
        self.assertEqual(
            instance.P.extract_values(),
            {('A3', 'B3'): 4.5, ('A1', 'B1'): 4.3, ('A2', 'B2'): 4.4},
        )
        self.assertEqual(
            instance.O.extract_values(),
            {('A3', 'B3'): 5.5, ('A1', 'B1'): 5.3, ('A2', 'B2'): 5.4},
        )

    def test_store_set1(self):
        # Write 1-D set
        self.check_skiplist('store_set1')
        model = ConcreteModel()
        model.A = Set(initialize=[1, 3, 5])
        data = DataPortal()
        data.store(set=model.A, **self.create_write_options('set1'))
        self.compare_data('set1', self.suffix)

    def test_store_set2(self):
        # Write 2-D set
        self.check_skiplist('store_set2')
        model = ConcreteModel()
        model.A = Set(initialize=[(1, 2), (3, 4), (5, 6)], dimen=2)
        data = DataPortal()
        data.store(set=model.A, **self.create_write_options('set2'))
        self.compare_data('set2', self.suffix)

    def test_store_param1(self):
        # Write scalar param
        self.check_skiplist('store_param1')
        model = ConcreteModel()
        model.p = Param(initialize=1)
        data = DataPortal()
        data.store(param=model.p, **self.create_write_options('param1'))
        self.compare_data('param1', self.suffix)

    def test_store_param2(self):
        # Write 1-D param
        self.check_skiplist('store_param2')
        model = ConcreteModel()
        model.A = Set(initialize=[1, 2, 3])
        model.p = Param(model.A, initialize={1: 10, 2: 20, 3: 30})
        data = DataPortal()
        data.store(param=model.p, **self.create_write_options('param2'))
        self.compare_data('param2', self.suffix)

    def test_store_param3(self):
        # Write 2-D params
        self.check_skiplist('store_param3')
        model = ConcreteModel()
        model.A = Set(initialize=[(1, 2), (2, 3), (3, 4)], dimen=2)
        model.p = Param(model.A, initialize={(1, 2): 10, (2, 3): 20, (3, 4): 30})
        model.q = Param(model.A, initialize={(1, 2): 11, (2, 3): 21, (3, 4): 31})
        data = DataPortal()
        data.store(param=(model.p, model.q), **self.create_write_options('param3'))
        self.compare_data('param3', self.suffix)

    def test_store_param4(self):
        # Write 2-D params
        self.check_skiplist('store_param4')
        model = ConcreteModel()
        model.A = Set(initialize=[(1, 2), (2, 3), (3, 4)], dimen=2)
        model.p = Param(model.A, initialize={(1, 2): 10, (2, 3): 20, (3, 4): 30})
        model.q = Param(model.A, initialize={(1, 2): 11, (2, 3): 21, (3, 4): 31})
        data = DataPortal()
        data.store(param=(model.p, model.q), **self.create_write_options('param4'))
        self.compare_data('param4', self.suffix)


class TestCsvPortal(TestTextPortal):
    suffix = '.csv'

    def create_options(self, name):
        return {
            'filename': os.path.abspath(
                tutorial_dir + os.sep + 'csv' + os.sep + name + self.suffix
            )
        }


class TestXmlPortal(TestTextPortal):
    suffix = '.xml'
    skiplist = ['tableD', 'tableT', 'tableU']

    def create_options(self, name):
        return {
            'filename': os.path.abspath(
                tutorial_dir + os.sep + 'xml' + os.sep + name + self.suffix
            )
        }


class TestJsonPortal(TestTextPortal):
    suffix = '.json'
    skiplist = ['tableD', 'tableT', 'tableU', 'tableXW_4']

    def create_options(self, name):
        return {
            'filename': os.path.abspath(
                tutorial_dir + os.sep + 'json' + os.sep + name + self.suffix
            )
        }


@unittest.skipIf(not yaml_interface, "YAML interface not available")
class TestYamlPortal(TestTextPortal):
    suffix = '.yaml'
    skiplist = ['tableD', 'tableT', 'tableU', 'tableXW_4']

    def create_options(self, name):
        return {
            'filename': os.path.abspath(
                tutorial_dir + os.sep + 'yaml' + os.sep + name + self.suffix
            )
        }


class LoadTests:
    skiplist = []

    #
    # FIXME: partially reverting 704a4de2.  That commit re-enabled these
    # tests after they had been disabled prior to the conversion of
    # Coopr to Pyomo.  However, these LoadTests are failing (and
    # presumably have been for 3+ years). This is disrupting ongoing
    # development, as it is masking new test results.
    #
    # For the moment, we will go back to sticking our head in the sand.
    #
    def check_skiplist(self, name):
        self.skipTest('Skipping load tests')

    def Xcheck_skiplist(self, name):
        if name in self.skiplist:
            self.skipTest('Skipping test %s' % name)

    def filename(self, tname):
        return os.path.abspath(
            tutorial_dir + os.sep + self.suffix + os.sep + tname + '.' + self.suffix
        )

    def test_tableA1(self):
        # Importing a single column of data
        self.check_skiplist('tableA1')
        with capture_output(currdir + 'loadA1.dat'):
            print("load " + self.filename('A') + " format=set : A;")
        model = AbstractModel()
        model.A = Set()
        instance = model.create_instance(currdir + 'loadA1.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3']))
        os.remove(currdir + 'loadA1.dat')

    def test_tableA2(self):
        # Importing a single column of data
        self.check_skiplist('tableA2')
        with capture_output(currdir + 'loadA2.dat'):
            print("load " + self.filename('A') + " ;")
        model = AbstractModel()
        model.A = Set()
        try:
            instance = model.create_instance(currdir + 'loadA2.dat')
            self.fail("Should fail because no set name is specified")
        except IOError:
            pass
        except IndexError:
            pass
        os.remove(currdir + 'loadA2.dat')

    def test_tableA3(self):
        # Importing a single column of data
        self.check_skiplist('tableA3')
        with capture_output(currdir + 'loadA3.dat'):
            print("load " + self.filename('A') + " format=set : A ;")
        model = AbstractModel()
        model.A = Set()
        instance = model.create_instance(currdir + 'loadA3.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3']))
        os.remove(currdir + 'loadA3.dat')

    def test_tableB1(self):
        # Same as test_tableA
        self.check_skiplist('tableB1')
        with capture_output(currdir + 'loadB.dat'):
            print("load " + self.filename('B') + " format=set : B;")
        model = AbstractModel()
        model.B = Set()
        instance = model.create_instance(currdir + 'loadB.dat')
        self.assertEqual(set(instance.B.data()), set([1, 2, 3]))
        os.remove(currdir + 'loadB.dat')

    def test_tableC(self):
        # Importing a multi-column table, where all columns are
        # treated as values for a set with tuple values.
        self.check_skiplist('tableC')
        with capture_output(currdir + 'loadC.dat'):
            print("load " + self.filename('C') + " format=set : C ;")
        model = AbstractModel()
        model.C = Set(dimen=2)
        instance = model.create_instance(currdir + 'loadC.dat')
        self.assertEqual(
            set(instance.C.data()),
            set(
                [
                    ('A1', 1),
                    ('A1', 2),
                    ('A1', 3),
                    ('A2', 1),
                    ('A2', 2),
                    ('A2', 3),
                    ('A3', 1),
                    ('A3', 2),
                    ('A3', 3),
                ]
            ),
        )
        os.remove(currdir + 'loadC.dat')

    def test_tableD(self):
        # Importing a 2D array of data as a set.
        self.check_skiplist('tableD')
        with capture_output(currdir + 'loadD.dat'):
            print("load " + self.filename('D') + " format=set_array : C ;")
        model = AbstractModel()
        model.C = Set(dimen=2)
        instance = model.create_instance(currdir + 'loadD.dat')
        self.assertEqual(set(instance.C.data()), set([('A1', 1), ('A2', 2), ('A3', 3)]))
        os.remove(currdir + 'loadD.dat')

    def test_tableZ(self):
        # Importing a single parameter
        self.check_skiplist('tableZ')
        with capture_output(currdir + 'loadZ.dat'):
            print("load " + self.filename('Z') + " : Z ;")
        model = AbstractModel()
        model.Z = Param(default=99.0)
        instance = model.create_instance(currdir + 'loadZ.dat')
        self.assertEqual(instance.Z, 1.01)
        os.remove(currdir + 'loadZ.dat')

    def test_tableY(self):
        # Same as tableXW.
        self.check_skiplist('tableY')
        with capture_output(currdir + 'loadY.dat'):
            print("load " + self.filename('Y') + " : [A] Y;")
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3', 'A4'])
        model.Y = Param(model.A)
        instance = model.create_instance(currdir + 'loadY.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3', 'A4']))
        self.assertEqual(instance.Y.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        os.remove(currdir + 'loadY.dat')

    def test_tableXW_1(self):
        # Importing a table, but only reporting the values for the non-index
        # parameter columns.  The first column is assumed to represent an
        # index column.
        self.check_skiplist('tableXW_1')
        with capture_output(currdir + 'loadXW.dat'):
            print("load " + self.filename('XW') + " : [A] X W;")
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3', 'A4'])
        model.X = Param(model.A)
        model.W = Param(model.A)
        instance = model.create_instance(currdir + 'loadXW.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3', 'A4']))
        self.assertEqual(instance.X.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(instance.W.extract_values(), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})
        os.remove(currdir + 'loadXW.dat')

    def test_tableXW_2(self):
        # Like test_tableXW_1, except that set A is not defined.
        self.check_skiplist('tableXW_2')
        with capture_output(currdir + 'loadXW.dat'):
            print("load " + self.filename('XW') + " : [A] X W;")
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3'])
        model.X = Param(model.A)
        model.W = Param(model.A)
        instance = model.create_instance(currdir + 'loadXW.dat')
        self.assertEqual(instance.X.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(instance.W.extract_values(), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})
        os.remove(currdir + 'loadXW.dat')

    def test_tableXW_3(self):
        # Like test_tableXW_1, except that set A is defined in the load statement.
        self.check_skiplist('tableXW_3')
        with capture_output(currdir + 'loadXW.dat'):
            print("load " + self.filename('XW') + " : A=[A] X W;")
        model = AbstractModel()
        model.A = Set()
        model.X = Param(model.A)
        model.W = Param(model.A)
        instance = model.create_instance(currdir + 'loadXW.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3']))
        self.assertEqual(instance.X.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(instance.W.extract_values(), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})
        os.remove(currdir + 'loadXW.dat')

    def test_tableXW_4(self):
        # Like test_tableXW_1, except that set A is defined in the load statement and all values are mapped.
        self.check_skiplist('tableXW_4')
        with capture_output(currdir + 'loadXW.dat'):
            print("load " + self.filename('XW') + " : B=[A] R=X S=W;")
        model = AbstractModel()
        model.B = Set()
        model.R = Param(model.B)
        model.S = Param(model.B)
        instance = model.create_instance(currdir + 'loadXW.dat')
        self.assertEqual(set(instance.B.data()), set(['A1', 'A2', 'A3']))
        self.assertEqual(instance.R.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(instance.S.extract_values(), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})
        os.remove(currdir + 'loadXW.dat')

    def test_tableT(self):
        # Importing a 2D array of parameters that are transposed.
        self.check_skiplist('tableT')
        with capture_output(currdir + 'loadT.dat'):
            print("load " + self.filename('T') + " format=transposed_array : T;")
        model = AbstractModel()
        model.B = Set(initialize=['I1', 'I2', 'I3', 'I4'])
        model.A = Set(initialize=['A1', 'A2', 'A3'])
        model.T = Param(model.A, model.B)
        instance = model.create_instance(currdir + 'loadT.dat')
        self.assertEqual(
            instance.T.extract_values(),
            {
                ('A2', 'I1'): 2.3,
                ('A1', 'I2'): 1.4,
                ('A1', 'I3'): 1.5,
                ('A1', 'I4'): 1.6,
                ('A1', 'I1'): 1.3,
                ('A3', 'I4'): 3.6,
                ('A2', 'I4'): 2.6,
                ('A3', 'I1'): 3.3,
                ('A2', 'I3'): 2.5,
                ('A3', 'I2'): 3.4,
                ('A2', 'I2'): 2.4,
                ('A3', 'I3'): 3.5,
            },
        )
        os.remove(currdir + 'loadT.dat')

    def test_tableU(self):
        # Importing a 2D array of parameters.
        self.check_skiplist('tableU')
        with capture_output(currdir + 'loadU.dat'):
            print("load " + self.filename('U') + " format=array : U;")
        model = AbstractModel()
        model.A = Set(initialize=['I1', 'I2', 'I3', 'I4'])
        model.B = Set(initialize=['A1', 'A2', 'A3'])
        model.U = Param(model.A, model.B)
        instance = model.create_instance(currdir + 'loadU.dat')
        self.assertEqual(
            instance.U.extract_values(),
            {
                ('I2', 'A1'): 1.4,
                ('I3', 'A1'): 1.5,
                ('I3', 'A2'): 2.5,
                ('I4', 'A1'): 1.6,
                ('I3', 'A3'): 3.5,
                ('I1', 'A2'): 2.3,
                ('I4', 'A3'): 3.6,
                ('I1', 'A3'): 3.3,
                ('I4', 'A2'): 2.6,
                ('I2', 'A3'): 3.4,
                ('I1', 'A1'): 1.3,
                ('I2', 'A2'): 2.4,
            },
        )
        os.remove(currdir + 'loadU.dat')

    def test_tableS(self):
        # Importing a table, but only reporting the values for the non-index
        # parameter columns.  The first column is assumed to represent an
        # index column.  A missing value is represented in the column data.
        self.check_skiplist('tableS')
        with capture_output(currdir + 'loadS.dat'):
            print("load " + self.filename('S') + " : [A] S ;")
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3', 'A4'])
        model.S = Param(model.A)
        instance = model.create_instance(currdir + 'loadS.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3', 'A4']))
        self.assertEqual(instance.S.extract_values(), {'A1': 3.3, 'A3': 3.5})
        os.remove(currdir + 'loadS.dat')

    def test_tablePO(self):
        # Importing a table that has multiple indexing columns
        self.check_skiplist('tablePO')
        with capture_output(currdir + 'loadPO.dat'):
            print("load " + self.filename('PO') + " : J=[A,B] P O;")
        model = AbstractModel()
        model.J = Set(dimen=2)
        model.P = Param(model.J)
        model.O = Param(model.J)
        instance = model.create_instance(currdir + 'loadPO.dat')
        self.assertEqual(
            set(instance.J.data()), set([('A3', 'B3'), ('A1', 'B1'), ('A2', 'B2')])
        )
        self.assertEqual(
            instance.P.extract_values(),
            {('A3', 'B3'): 4.5, ('A1', 'B1'): 4.3, ('A2', 'B2'): 4.4},
        )
        self.assertEqual(
            instance.O.extract_values(),
            {('A3', 'B3'): 5.5, ('A1', 'B1'): 5.3, ('A2', 'B2'): 5.4},
        )
        os.remove(currdir + 'loadPO.dat')


class TestTextLoad(LoadTests, unittest.TestCase):
    suffix = 'tab'


class TestCsvLoad(LoadTests, unittest.TestCase):
    suffix = 'csv'


class TestXmlLoad(LoadTests, unittest.TestCase):
    suffix = 'xml'
    skiplist = ['tableD', 'tableT', 'tableU']

    def test_tableXW_nested1(self):
        # Importing a table, but only reporting the values for the non-index
        # parameter columns.  The first column is assumed to represent an
        # index column.
        self.check_skiplist('tableXW_1')
        with capture_output(currdir + 'loadXW.dat'):
            print(
                "load "
                + self.filename('XW_nested1')
                + " query='./bar/table/*' : [A] X W;"
            )
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3', 'A4'])
        model.X = Param(model.A)
        model.W = Param(model.A)
        instance = model.create_instance(currdir + 'loadXW.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3', 'A4']))
        self.assertEqual(instance.X.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(instance.W.extract_values(), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})
        os.remove(currdir + 'loadXW.dat')

    def test_tableXW_nested2(self):
        # Importing a table, but only reporting the values for the non-index
        # parameter columns.  The first column is assumed to represent an
        # index column.
        self.check_skiplist('tableXW_1')
        with capture_output(currdir + 'loadXW.dat'):
            print(
                "load "
                + self.filename('XW_nested2')
                + " query='./bar/table/row' : [A] X W;"
            )
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3', 'A4'])
        model.X = Param(model.A)
        model.W = Param(model.A)
        instance = model.create_instance(currdir + 'loadXW.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3', 'A4']))
        self.assertEqual(instance.X.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(instance.W.extract_values(), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})
        os.remove(currdir + 'loadXW.dat')


class Spreadsheet(LoadTests):
    def filename(self, tname):
        if tname == "Z":
            return (
                os.path.abspath(tutorial_dir + os.sep + self._filename)
                + " range="
                + tname
                + "param"
            )
        else:
            return (
                os.path.abspath(tutorial_dir + os.sep + self._filename)
                + " range="
                + tname
                + "table"
            )


@unittest.skipIf(not xls_interface, "No XLS interface available")
class TestSpreadsheetXLS(Spreadsheet, unittest.TestCase):
    _filename = 'excel.xls'


@unittest.skipIf(not xlsx_interface, "No XLSX interface available")
class TestSpreadsheetXLSX(Spreadsheet, unittest.TestCase):
    _filename = 'excel.xlsx'


@unittest.skipIf(not xlsb_interface, "No XLSB interface available")
class TestSpreadsheetXLSB(Spreadsheet, unittest.TestCase):
    _filename = 'excel.xlsb'


@unittest.skipIf(not xlsm_interface, "No XLSM interface available")
class TestSpreadsheetXLSM(Spreadsheet, unittest.TestCase):
    _filename = 'excel.xlsm'


class TestTableCmd(unittest.TestCase):
    def test_tableA1_1(self):
        # Importing a single column of data as a set
        with capture_output(currdir + 'loadA1.dat'):
            print("table columns=1 A={1} := A1 A2 A3 ;")
        model = AbstractModel()
        model.A = Set()
        instance = model.create_instance(currdir + 'loadA1.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3']))
        os.remove(currdir + 'loadA1.dat')

    def test_tableA1_2(self):
        # Importing a single column of data as a set
        with capture_output(currdir + 'loadA1.dat'):
            print("table A={A} : A := A1 A2 A3 ;")
        model = AbstractModel()
        model.A = Set()
        instance = model.create_instance(currdir + 'loadA1.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3']))
        os.remove(currdir + 'loadA1.dat')

    def test_tableB1_1(self):
        # Same as test_tableA
        with capture_output(currdir + 'loadB.dat'):
            print("table columns=1 B={1} := 1 2 3 ;")
        model = AbstractModel()
        model.B = Set()
        instance = model.create_instance(currdir + 'loadB.dat')
        self.assertEqual(set(instance.B.data()), set([1, 2, 3]))
        os.remove(currdir + 'loadB.dat')

    def test_tableB1_2(self):
        # Same as test_tableA
        with capture_output(currdir + 'loadB.dat'):
            print("table B={B} : B := 1 2 3 ;")
        model = AbstractModel()
        model.B = Set()
        instance = model.create_instance(currdir + 'loadB.dat')
        self.assertEqual(set(instance.B.data()), set([1, 2, 3]))
        os.remove(currdir + 'loadB.dat')

    def test_tableC_1(self):
        # Importing a multi-column table, where all columns are
        # treated as values for a set with tuple values.
        with capture_output(currdir + 'loadC.dat'):
            print(
                "table columns=2 C={1,2} := A1 1 A1 2 A1 3 A2 1 A2 2 A2 3 A3 1 A3 2 A3 3 ;"
            )
        model = AbstractModel()
        model.C = Set(dimen=2)
        instance = model.create_instance(currdir + 'loadC.dat')
        self.assertEqual(
            set(instance.C.data()),
            set(
                [
                    ('A1', 1),
                    ('A1', 2),
                    ('A1', 3),
                    ('A2', 1),
                    ('A2', 2),
                    ('A2', 3),
                    ('A3', 1),
                    ('A3', 2),
                    ('A3', 3),
                ]
            ),
        )
        os.remove(currdir + 'loadC.dat')

    def test_tableC_2(self):
        # Importing a multi-column table, where all columns are
        # treated as values for a set with tuple values.
        with capture_output(currdir + 'loadC.dat'):
            print(
                "table C={a,b} : a b := A1 1 A1 2 A1 3 A2 1 A2 2 A2 3 A3 1 A3 2 A3 3 ;"
            )
        model = AbstractModel()
        model.C = Set(dimen=2)
        instance = model.create_instance(currdir + 'loadC.dat')
        self.assertEqual(
            set(instance.C.data()),
            set(
                [
                    ('A1', 1),
                    ('A1', 2),
                    ('A1', 3),
                    ('A2', 1),
                    ('A2', 2),
                    ('A2', 3),
                    ('A3', 1),
                    ('A3', 2),
                    ('A3', 3),
                ]
            ),
        )
        os.remove(currdir + 'loadC.dat')

    def test_tableZ(self):
        # Importing a single parameter
        with capture_output(currdir + 'loadZ.dat'):
            print("table Z := 1.01 ;")
        model = AbstractModel()
        model.Z = Param(default=99.0)
        instance = model.create_instance(currdir + 'loadZ.dat')
        self.assertEqual(instance.Z, 1.01)
        os.remove(currdir + 'loadZ.dat')

    def test_tableY_1(self):
        # Same as tableXW.
        with capture_output(currdir + 'loadY.dat'):
            print("table columns=2 Y(1)={2} := A1 3.3 A2 3.4 A3 3.5 ;")
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3', 'A4'])
        model.Y = Param(model.A)
        instance = model.create_instance(currdir + 'loadY.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3', 'A4']))
        self.assertEqual(instance.Y.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        os.remove(currdir + 'loadY.dat')

    def test_tableY_2(self):
        # Same as tableXW.
        with capture_output(currdir + 'loadY.dat'):
            print("table Y(A) : A Y := A1 3.3 A2 3.4 A3 3.5 ;")
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3', 'A4'])
        model.Y = Param(model.A)
        instance = model.create_instance(currdir + 'loadY.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3', 'A4']))
        self.assertEqual(instance.Y.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        os.remove(currdir + 'loadY.dat')

    def test_tableXW_1_1(self):
        # Importing a table, but only reporting the values for the non-index
        # parameter columns.  The first column is assumed to represent an
        # index column.
        with capture_output(currdir + 'loadXW.dat'):
            print(
                "table columns=3 X(1)={2} W(1)={3} := A1 3.3 4.3 A2 3.4 4.4 A3 3.5 4.5 ;"
            )
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3', 'A4'])
        model.X = Param(model.A)
        model.W = Param(model.A)
        instance = model.create_instance(currdir + 'loadXW.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3', 'A4']))
        self.assertEqual(instance.X.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(instance.W.extract_values(), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})
        os.remove(currdir + 'loadXW.dat')

    def test_tableXW_1_2(self):
        # Importing a table, but only reporting the values for the non-index
        # parameter columns.  The first column is assumed to represent an
        # index column.
        with capture_output(currdir + 'loadXW.dat'):
            print("table X(A) W(A) : A X W := A1 3.3 4.3 A2 3.4 4.4 A3 3.5 4.5 ;")
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3', 'A4'])
        model.X = Param(model.A)
        model.W = Param(model.A)
        instance = model.create_instance(currdir + 'loadXW.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3', 'A4']))
        self.assertEqual(instance.X.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(instance.W.extract_values(), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})
        os.remove(currdir + 'loadXW.dat')

    def test_tableXW_3_1(self):
        # Like test_tableXW_1, except that set A is defined in the load statement.
        with capture_output(currdir + 'loadXW.dat'):
            print(
                "table columns=3 A={1} X(A)={2} W(A)={3} := A1 3.3 4.3 A2 3.4 4.4 A3 3.5 4.5 ;"
            )
        model = AbstractModel()
        model.A = Set()
        model.X = Param(model.A)
        model.W = Param(model.A)
        instance = model.create_instance(currdir + 'loadXW.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3']))
        self.assertEqual(instance.X.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(instance.W.extract_values(), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})
        os.remove(currdir + 'loadXW.dat')

    def test_tableXW_3_2(self):
        # Like test_tableXW_1, except that set A is defined in the load statement.
        with capture_output(currdir + 'loadXW.dat'):
            print("table A={A} X(A) W(A) : A X W := A1 3.3 4.3 A2 3.4 4.4 A3 3.5 4.5 ;")
        model = AbstractModel()
        model.A = Set()
        model.X = Param(model.A)
        model.W = Param(model.A)
        instance = model.create_instance(currdir + 'loadXW.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3']))
        self.assertEqual(instance.X.extract_values(), {'A1': 3.3, 'A2': 3.4, 'A3': 3.5})
        self.assertEqual(instance.W.extract_values(), {'A1': 4.3, 'A2': 4.4, 'A3': 4.5})
        os.remove(currdir + 'loadXW.dat')

    def test_tableS_1(self):
        # Importing a table, but only reporting the values for the non-index
        # parameter columns.  The first column is assumed to represent an
        # index column.  A missing value is represented in the column data.
        with capture_output(currdir + 'loadS.dat'):
            print("table columns=2 S(1)={2} := A1 3.3 A2 . A3 3.5 ;")
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3', 'A4'])
        model.S = Param(model.A)
        instance = model.create_instance(currdir + 'loadS.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3', 'A4']))
        self.assertEqual(instance.S.extract_values(), {'A1': 3.3, 'A3': 3.5})
        os.remove(currdir + 'loadS.dat')

    def test_tableS_2(self):
        # Importing a table, but only reporting the values for the non-index
        # parameter columns.  The first column is assumed to represent an
        # index column.  A missing value is represented in the column data.
        with capture_output(currdir + 'loadS.dat'):
            print("table S(A) : A S := A1 3.3 A2 . A3 3.5 ;")
        model = AbstractModel()
        model.A = Set(initialize=['A1', 'A2', 'A3', 'A4'])
        model.S = Param(model.A)
        instance = model.create_instance(currdir + 'loadS.dat')
        self.assertEqual(set(instance.A.data()), set(['A1', 'A2', 'A3', 'A4']))
        self.assertEqual(instance.S.extract_values(), {'A1': 3.3, 'A3': 3.5})
        os.remove(currdir + 'loadS.dat')

    def test_tablePO_1(self):
        # Importing a table that has multiple indexing columns
        with capture_output(currdir + 'loadPO.dat'):
            print(
                "table columns=4 J={1,2} P(J)={3} O(J)={4} := A1 B1 4.3 5.3 A2 B2 4.4 5.4 A3 B3 4.5 5.5 ;"
            )
        model = AbstractModel()
        model.J = Set(dimen=2)
        model.P = Param(model.J)
        model.O = Param(model.J)
        instance = model.create_instance(currdir + 'loadPO.dat')
        self.assertEqual(
            set(instance.J.data()), set([('A3', 'B3'), ('A1', 'B1'), ('A2', 'B2')])
        )
        self.assertEqual(
            instance.P.extract_values(),
            {('A3', 'B3'): 4.5, ('A1', 'B1'): 4.3, ('A2', 'B2'): 4.4},
        )
        self.assertEqual(
            instance.O.extract_values(),
            {('A3', 'B3'): 5.5, ('A1', 'B1'): 5.3, ('A2', 'B2'): 5.4},
        )
        os.remove(currdir + 'loadPO.dat')

    def test_tablePO_2(self):
        # Importing a table that has multiple indexing columns
        with capture_output(currdir + 'loadPO.dat'):
            print(
                "table J={A,B} P(J) O(J) : A B P O := A1 B1 4.3 5.3 A2 B2 4.4 5.4 A3 B3 4.5 5.5 ;"
            )
        model = AbstractModel()
        model.J = Set(dimen=2)
        model.P = Param(model.J)
        model.O = Param(model.J)
        instance = model.create_instance(currdir + 'loadPO.dat')
        self.assertEqual(
            set(instance.J.data()), set([('A3', 'B3'), ('A1', 'B1'), ('A2', 'B2')])
        )
        self.assertEqual(
            instance.P.extract_values(),
            {('A3', 'B3'): 4.5, ('A1', 'B1'): 4.3, ('A2', 'B2'): 4.4},
        )
        self.assertEqual(
            instance.O.extract_values(),
            {('A3', 'B3'): 5.5, ('A1', 'B1'): 5.3, ('A2', 'B2'): 5.4},
        )
        os.remove(currdir + 'loadPO.dat')

    def test_complex_1(self):
        # Importing a table with multiple indexing columns
        with capture_output(currdir + 'loadComplex.dat'):
            print("table columns=8 I={4} J={3,5} A(I)={1} B(J)={7} :=")
            print("A1 x1 J311 I1 J321 y1 B1 z1")
            print("A2 x2 J312 I2 J322 y2 B2 z2")
            print("A3 x3 J313 I3 J323 y3 B3 z3")
            print(";")
        model = AbstractModel()
        model.I = Set()
        model.J = Set(dimen=2)
        model.A = Param(model.I)
        model.B = Param(model.J)
        instance = model.create_instance(currdir + 'loadComplex.dat')
        self.assertEqual(
            set(instance.J.data()),
            set([('J311', 'J321'), ('J312', 'J322'), ('J313', 'J323')]),
        )
        self.assertEqual(set(instance.I.data()), set(['I1', 'I2', 'I3']))
        self.assertEqual(
            instance.B.extract_values(),
            {('J311', 'J321'): 'B1', ('J312', 'J322'): 'B2', ('J313', 'J323'): 'B3'},
        )
        self.assertEqual(
            instance.A.extract_values(), {'I1': 'A1', 'I2': 'A2', 'I3': 'A3'}
        )
        os.remove(currdir + 'loadComplex.dat')

    def test_complex_2(self):
        # Importing a table with multiple indexing columns
        with capture_output(currdir + 'loadComplex.dat'):
            print("table I={I} J={J1,J2} A(J) B(I) :")
            print("A  x  J1   I  J2   y  B  z :=")
            print("A1 x1 J311 I1 J321 y1 B1 z1")
            print("A2 x2 J312 I2 J322 y2 B2 z2")
            print("A3 x3 J313 I3 J323 y3 B3 z3")
            print(";")
        model = AbstractModel()
        model.I = Set()
        model.J = Set(dimen=2)
        model.A = Param(model.J)
        model.B = Param(model.I)
        instance = model.create_instance(currdir + 'loadComplex.dat')
        self.assertEqual(
            set(instance.J.data()),
            set([('J311', 'J321'), ('J312', 'J322'), ('J313', 'J323')]),
        )
        self.assertEqual(set(instance.I.data()), set(['I1', 'I2', 'I3']))
        self.assertEqual(
            instance.A.extract_values(),
            {('J311', 'J321'): 'A1', ('J312', 'J322'): 'A2', ('J313', 'J323'): 'A3'},
        )
        self.assertEqual(
            instance.B.extract_values(), {'I1': 'B1', 'I2': 'B2', 'I3': 'B3'}
        )
        os.remove(currdir + 'loadComplex.dat')


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.common.dependencies import networkx as nx, networkx_available
from pyomo.contrib.incidence_analysis.common.dulmage_mendelsohn import (
    dulmage_mendelsohn,
)


@unittest.skipUnless(networkx_available, "networkx is not available")
class TestPothenFanExample(unittest.TestCase):
    """
    Test on the first example graph from Pothen and Fan, 1990.
    """

    def _construct_graph(self):
        bg = nx.Graph()
        n_l = 12
        n_r = 11
        left_nodes = list(range(n_l))
        right_nodes = list(range(n_r))

        bg.add_nodes_from(left_nodes, bipartite=0)
        bg.add_nodes_from([n_l + i for i in right_nodes], bipartite=1)

        paper_edges = [
            (1, 1),
            (1, 2),
            (1, 3),
            (1, 4),
            (1, 5),
            (1, 6),
            (2, 4),
            (2, 5),
            (2, 7),
            (2, 8),
            (2, 10),
            (3, 1),
            (3, 3),
            (3, 5),
            (4, 6),
            (4, 7),
            (4, 11),
            (5, 6),
            (5, 7),
            (5, 9),
            (6, 8),
            (6, 9),
            (7, 8),
            (7, 9),
            (7, 10),
            (8, 10),
            (8, 11),
            (9, 11),
            (10, 10),
            (11, 10),
            (11, 11),
            (12, 11),
        ]

        edges = [(i - 1, j - 1 + n_l) for i, j in paper_edges]
        bg.add_edges_from(edges)

        return bg

    def test_coarse_partition(self):
        bg = self._construct_graph()
        left_partition, right_partition = dulmage_mendelsohn(bg)

        # As partitioned in the paper

        # Potentially unmatched
        left_underconstrained = {7, 8, 9, 10, 11}
        # Matched with potentially unmatched left nodes
        right_overconstrained = {21, 22}

        right_underconstrained = {12, 13, 14, 15, 16}
        left_overconstrained = {0, 1, 2}

        left_square = {3, 4, 5, 6}
        right_square = {17, 18, 19, 20}

        nodes = left_partition[0] + left_partition[1]
        self.assertEqual(set(nodes), left_underconstrained)

        nodes = right_partition[2]
        self.assertEqual(set(nodes), right_overconstrained)

        nodes = right_partition[0] + right_partition[1]
        self.assertEqual(set(nodes), right_underconstrained)

        nodes = left_partition[2]
        self.assertEqual(set(nodes), left_overconstrained)

        nodes = left_partition[3]
        self.assertEqual(set(nodes), left_square)

        nodes = right_partition[3]
        self.assertEqual(set(nodes), right_square)


@unittest.skipUnless(networkx_available, "networkx is not available")
class TestDisconnectedModel(unittest.TestCase):
    """
    An error is raised if top_nodes are not provided for a disconnected
    graph, which we test here.
    """

    def _construct_graph(self):
        """
        Graph with the following incidence matrix:
        |x x         x|
        |x   x        |
        |      x   x  |
        |        x x  |
        |      x x    |
        |            x|
        |            x|
        """
        N = 7
        top_nodes = list(range(N))
        bot_nodes = list(range(N, 2 * N))

        graph = nx.Graph()
        graph.add_nodes_from(top_nodes, bipartite=0)
        graph.add_nodes_from(bot_nodes, bipartite=1)

        edges = [
            (0, 0),
            (0, 1),
            (0, 6),
            (1, 0),
            (1, 2),
            (2, 3),
            (2, 5),
            (3, 4),
            (3, 5),
            (4, 3),
            (4, 4),
            (5, 6),
            (6, 6),
        ]
        edges = [(i, j + N) for i, j in edges]
        graph.add_edges_from(edges)
        return graph, top_nodes

    def test_graph_dm_partition(self):
        graph, top_nodes = self._construct_graph()

        with self.assertRaises(nx.AmbiguousSolution):
            top_dmp, bot_dmp = dulmage_mendelsohn(graph)

        top_dmp, bot_dmp = dulmage_mendelsohn(graph, top_nodes=top_nodes)

        self.assertFalse(nx.is_connected(graph))

        underconstrained_top = {0, 1}
        underconstrained_bot = {7, 8, 9}
        self.assertEqual(underconstrained_top, set(top_dmp[2]))
        self.assertEqual(underconstrained_bot, set(bot_dmp[0] + bot_dmp[1]))

        overconstrained_top = {5, 6}
        overconstrained_bot = {13}
        self.assertEqual(overconstrained_top, set(top_dmp[0] + top_dmp[1]))
        self.assertEqual(overconstrained_bot, set(bot_dmp[2]))

        wellconstrained_top = {2, 3, 4}
        wellconstrained_bot = {10, 11, 12}
        self.assertEqual(wellconstrained_top, set(top_dmp[3]))
        self.assertEqual(wellconstrained_bot, set(bot_dmp[3]))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
from pyomo.common.collections import ComponentMap
from pyomo.gdp.util import clone_without_expression_components
from pyomo.contrib.fbbt.fbbt import compute_bounds_on_expr
from pyomo.contrib.alternative_solutions import aos_utils


def _get_unique_name(collection, name):
    """Create a unique name for an item that will be added to a collection."""
    if name not in collection:
        return name
    else:
        i = 1
        while "{}_{}".format(name, i) not in collection:
            i += 1
        return "{}_{}".format(name, i)


def _set_slack_ub(expression, slack_var):
    """
    Use FBBT to compute an upper bound for a slack variable on an equality
    expression."""
    slack_lb, slack_ub = compute_bounds_on_expr(expression)
    assert slack_ub >= 0
    slack_var.setub(slack_ub)


def get_shifted_linear_model(model, block=None):
    """
    Converts an (MI)LP with bounded (discrete and) continuous variables
    (l <= x <= u) into a standard form where where all continuous variables
    are non-negative reals and all constraints are equalities. For a pure LP of
    the form,

    .. math::

       min/max cx
       s.t.
           A_1 * x  =  b_1
           A_2 * x <= b_2
           l <= x <= u

    a problem of the form,

    .. math::

       min/max c'z
       s.t.
           Bz = q
           z >= 0

    will be created and added to the returned block. z consists of var_lower
    and var_upper variables that are substituted into the original x variables,
    and slack_vars that are used to convert the original inequalities to
    equalities. Bounds are provided on all variables in z. For MILPs, only the
    continuous part of the problem is converted.

    See Lee, Sangbum., C. Phalakornkule, M. Domach, I. Grossmann, Recursive
    MILP model for finding all the alternate optima in LP models for metabolic
    networks, Computers & Chemical Engineering, Volume 24, Issues 2–7, 2000,
    page 712 for additional details.

    Parameters
    ----------
    model : ConcreteModel
        A concrete Pyomo model
    block : Block
        The Pyomo block that the new model should be added to.

    Returns
    -------
    block
        The block that holds the reformulated model.
    """

    # Gather all variables and confirm the model is bounded
    all_vars = aos_utils.get_model_variables(model)
    new_vars = {}
    all_vars_new = {}
    var_map = ComponentMap()
    var_range = {}
    for var in all_vars:
        assert var.lb is not None, (
            "Variable {} does not have a "
            "lower bound. All variables must be "
            "bounded.".format(var.name)
        )
        assert var.ub is not None, (
            "Variable {} does not have an "
            "upper bound. All variables must be "
            "bounded.".format(var.name)
        )
        if var.is_continuous():
            var_name = _get_unique_name(new_vars.keys(), var.name)
            new_vars[var_name] = var
            all_vars_new[var_name] = var
            var_map[var] = var_name
            var_range[var_name] = (0, var.ub - var.lb)
        else:
            all_vars_new[var.name] = var

    if block is None:
        block = model
    shifted_lp = aos_utils._add_aos_block(block, name="_shifted_lp")

    # Replace original variables with shifted lower and upper variables
    shifted_lp.var_lower = pyo.Var(
        new_vars.keys(), domain=pyo.NonNegativeReals, bounds=var_range
    )
    shifted_lp.var_upper = pyo.Var(
        new_vars.keys(), domain=pyo.NonNegativeReals, bounds=var_range
    )

    # Link the shifted lower and upper variables
    def link_vars_rule(m, var_index):
        return (
            m.var_lower[var_index] + m.var_upper[var_index] == m.var_upper[var_index].ub
        )

    shifted_lp.link_vars = pyo.Constraint(new_vars.keys(), rule=link_vars_rule)

    # Map the lower and upper variables to the original variables and their
    # lower bounds. This will be used to substitute x with var_lower + x.lb.
    var_lower_map = {id(var): shifted_lp.var_lower[i] for i, var in new_vars.items()}
    var_lower_bounds = {id(var): var.lb for var in new_vars.values()}
    var_zeros = {id(var): 0 for var in all_vars_new.values()}

    # Substitute the new s variables into the objective function
    # The c_fix_zeros calculation is used to find any constant terms that exist
    # in the objective expression to avoid double counting
    active_objective = aos_utils.get_active_objective(model)
    c_var_lower = clone_without_expression_components(
        active_objective.expr, substitute=var_lower_map
    )
    c_fix_lower = clone_without_expression_components(
        active_objective.expr, substitute=var_lower_bounds
    )
    c_fix_zeros = clone_without_expression_components(
        active_objective.expr, substitute=var_zeros
    )
    shifted_lp.objective = pyo.Objective(
        expr=c_var_lower - c_fix_zeros + c_fix_lower,
        name=active_objective.name + "_shifted",
        sense=active_objective.sense,
    )

    # Identify all of the shifted constraints and associated slack variables
    # that will need to be created
    new_constraints = {}
    constraint_map = ComponentMap()
    constraint_type = {}
    slacks = []
    for constraint in model.component_data_objects(pyo.Constraint, active=True):
        if constraint.parent_block() == shifted_lp:
            continue
        if constraint.equality:
            constraint_name = constraint.name + "_equal"
            constraint_name = _get_unique_name(new_constraints.keys(), constraint.name)
            new_constraints[constraint_name] = constraint
            constraint_map[constraint] = constraint_name
            constraint_type[constraint_name] = 0
        else:
            if constraint.lb is not None:
                constraint_name = constraint.name + "_lower"
                constraint_name = _get_unique_name(
                    new_constraints.keys(), constraint.name
                )
                new_constraints[constraint_name] = constraint
                constraint_map[constraint] = constraint_name
                constraint_type[constraint_name] = -1
                slacks.append(constraint_name)
            if constraint.ub is not None:
                constraint_name = constraint.name + "_upper"
                constraint_name = _get_unique_name(
                    new_constraints.keys(), constraint.name
                )
                new_constraints[constraint_name] = constraint
                constraint_map[constraint] = constraint_name
                constraint_type[constraint_name] = 1
                slacks.append(constraint_name)
    shifted_lp.constraint_index = pyo.Set(initialize=new_constraints.keys())
    shifted_lp.slack_index = pyo.Set(initialize=slacks)
    shifted_lp.slack_vars = pyo.Var(shifted_lp.slack_index, domain=pyo.NonNegativeReals)
    shifted_lp.constraints = pyo.Constraint(shifted_lp.constraint_index)

    for constraint_name, constraint in new_constraints.items():
        # The c_fix_zeros calculation is used to find any constant terms that
        # exist in the constraint expression to avoid double counting
        a_sub_var_lower = clone_without_expression_components(
            constraint.body, substitute=var_lower_map
        )
        a_sub_fix_lower = clone_without_expression_components(
            constraint.body, substitute=var_lower_bounds
        )
        a_sub_fix_zeros = clone_without_expression_components(
            constraint.body, substitute=var_zeros
        )
        b_lower = constraint.lb
        b_upper = constraint.ub
        con_type = constraint_type[constraint_name]
        if con_type == 0:
            expr = a_sub_var_lower - a_sub_fix_zeros + a_sub_fix_lower - b_lower == 0
        elif con_type == -1:
            expr_rhs = a_sub_var_lower - a_sub_fix_zeros + a_sub_fix_lower - b_lower
            expr = shifted_lp.slack_vars[constraint_name] == expr_rhs
            _set_slack_ub(expr_rhs, shifted_lp.slack_vars[constraint_name])
        elif con_type == 1:
            expr_rhs = b_upper - a_sub_var_lower + a_sub_fix_zeros - a_sub_fix_lower
            expr = shifted_lp.slack_vars[constraint_name] == expr_rhs
            _set_slack_ub(expr_rhs, shifted_lp.slack_vars[constraint_name])
        shifted_lp.constraints[constraint_name] = expr

    shifted_lp.var_map = var_map
    shifted_lp.new_vars = new_vars
    shifted_lp.constraint_map = constraint_map
    shifted_lp.new_constraints = new_constraints

    return shifted_lp
